/* callbacks.c - controls GTK callback functions (excluding network) */

#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>

#include <gnome.h>

#include <ctype.h>

#include "napster.h"
#include "gnapster.h"
#include "callbacks.h"
#include "ui.h"
#include "upload.h"
#include "network.h"
#include "commands.h"
#include "queue.h"
#include "download.h"
#include "servers.h"

#include "config.h"

extern CONN_INFO conn_info;
extern USER_INFO user_info;
extern GNAPSTER_MAIN gnapster_main;
extern SEARCH_TAB search_tab;
extern BROWSE_TAB browse_tab;
extern DOWNLOAD_TAB download_tab;
extern UPLOAD_TAB upload_tab;
extern CONSOLE_TAB console_tab;
extern MOTD_TAB motd_tab;
extern PROP_DLG prop_dlg;
extern BUILD_DLG build_dlg;
extern BROWSE_DLG browse_dlg;
extern ADD_DLG add_dlg;
extern EDIT_DLG edit_dlg;
extern LIST_CHAN_DLG list_chan_dlg;
extern RESUME_DLG resume_dlg;
extern REFRESH_DLG refresh_dlg;

extern int building;
extern GList *total_list;
extern GList *ignore_list;

#define MAX_HISTORY 15

int loc_curr = 0, loc_end = 0;
GList *console_history = NULL;

void gnapster_exit_cb() {
   if (conn_info.connected) {
      close(conn_info.sock); 
      gdk_input_remove(conn_info.sock_input);
   }
   gtk_main_quit();
}

void update_shares_cb() {
   shared_list_handle(user_info.upload_dir, UPDATE_LIST);   
}

void connect_last_cb() {
   if (conn_info.connected || conn_info.connecting)
     already_connected(NULL);
   else
     connect_cb();
}

void already_connected_cb(gint reply, gpointer data) {
   if (reply == GNOME_YES) {
      disconnect_cb();
      if (data) {
	 NAP_SERVER *nsrv;
	 
	 nsrv = (NAP_SERVER *)data;
	 browse_dlg_connect_cb(NULL, nsrv);
      } else
	connect_cb();
   } else if (reply != GNOME_NO) {
      g_assert_not_reached();
   } 
}

void already_connected(NAP_SERVER *nsrv) {
   GtkWidget *dialog;
   
   dialog = gnome_question_dialog_parented(_("You're already connected to a server. Disconnect?"),
					   already_connected_cb,
					   nsrv,
					   GTK_WINDOW(gnapster_main.window));
   
   gtk_widget_show(dialog);
}

void connect_official_cb() {
   gnome_config_set_string("/gnapster/Options/server", "208.184.216.222");
   gnome_config_set_int("/gnapster/Options/server_port", 8875);
   gnome_config_sync();
   if (conn_info.connected || conn_info.connecting) {
      already_connected(NULL);
//      gnome_error_dialog("You are already connected to a server!  Disconnect before attempting to use this operation");
      return;
   }
   connect_cb();
}

void switch_page_cb(GtkWidget *widget, GtkNotebookPage *page, int page_num,
		    gpointer data) {
   GtkWidget *nth_page, *label;
   GtkStyle *old_style;
   
   if (page_num == 4 || widget == console_tab.notebook)
     gtk_widget_grab_focus(console_tab.entry);
   
   nth_page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(widget), page_num);
   label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(widget), nth_page);
   old_style = gtk_widget_get_style(label);
   if (old_style == gnapster_main.highlight)
     gtk_widget_set_style(label, gnapster_main.normal);
}

void build_list_cb() {
   if (building) {
      building = 0;
      return;
   }

   shared_list_handle(gtk_entry_get_text(GTK_ENTRY(gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(prop_dlg.upload_dir)))), 
		      BUILD_LIST);
   if (conn_info.connected)
     gnapster_text_insert(CONSOLE, SYSTEM, "Please reconnect to the current server for the new mp3 list to take effect\n");
}

void update_list_cb() {
   shared_list_handle(gtk_entry_get_text(GTK_ENTRY(gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(prop_dlg.upload_dir)))),
		      UPDATE_LIST);
}

/*void list_chan_dlg_clist_cb(GtkWidget *widget, GdkEventButton *event, gpointer data) {
   int row = -1, col;
   
   gtk_clist_get_selection_info(GTK_CLIST(widget), event->x, event->y,
				&row, &col);
   
   if (event->type == GDK_2BUTTON_PRESS && event->button == 1) {
      list_chan_dlg_cb(NULL, "join");
   }
}

void list_chan_dlg_cb(GtkWidget *w, void *data) {
   if (!data) return;
   
   if (!strcmp(data, "join")) {
      NAP_CHANNEL *nchan;
      
      nchan = gnapster_get_selection_data(list_chan_dlg.clist, NULL);
      if (!nchan) return;
      napster_send(conn_info.sock, NAPSTER_JOIN_CHAN, nchan->chan);
   }
   
   gnome_dialog_close(GNOME_DIALOG(list_chan_dlg.window));
   list_chan_dlg.window = NULL;
}

void list_chan_cb(GtkWidget *w, gpointer data) {
   if (!conn_info.connected || conn_info.connecting) 
     return;
   create_list_chan_dlg();
   gtk_clist_freeze(GTK_CLIST(list_chan_dlg.clist));
   napster_send(conn_info.sock, NAPSTER_CHAN_LIST, NULL);
}

void join_chan_cb(GtkWidget *w, gpointer data) {
   gtk_notebook_set_page(GTK_NOTEBOOK(gnapster_main.notebook), 4);
   gtk_entry_set_text(GTK_ENTRY(console_tab.entry), "/join ");
   gtk_widget_grab_focus(console_tab.entry);
}

void part_chan_cb(GtkWidget *w, gpointer data) {
   if (!remove_console_tab(0))
     gnome_error_dialog("You must click on a channel tab before attempting to exit a channel");
}*/

void query_user_cb(GtkWidget *w, gpointer data) {
   gtk_notebook_set_page(GTK_NOTEBOOK(gnapster_main.notebook), 4);
   gtk_entry_set_text(GTK_ENTRY(console_tab.entry), "/query ");
   gtk_widget_grab_focus(console_tab.entry);
}

void unquery_user_cb(GtkWidget *w, gpointer data) {
   if (!remove_console_tab(1))
     gnome_error_dialog("You must click on a user query tab before attempting to close it");
}

void menu_popup_cb(GtkWidget *widget, GdkEventButton *event, gpointer data) {
   GtkWidget *temp_menu;
   int row = -1, col;

   gtk_clist_get_selection_info(GTK_CLIST(widget), event->x, event->y,
				&row, &col);

   if (event->type == GDK_2BUTTON_PRESS && event->button == 1 &&
       widget == search_tab.clist) {
      NAP_SEARCH *temp;
      NAP_DOWNLOAD *download;
      
      temp = gtk_clist_get_row_data(GTK_CLIST(widget), row);
      if (!temp) return;
      if (!temp->user) return;

      download = search_dup(temp, N_DOWNLOAD);
      handle_download(download, DOWNLOAD_NORMAL);
   } else if (event->type == GDK_2BUTTON_PRESS && event->button == 1 &&
	      widget == browse_tab.ctree) {
//      browse_download_cb(browse_tab.ctree, NULL);
/*      void *temp;
      GList *selection;
      NAP_DOWNLOAD *download;
      
      printf("widget == browse_tab.ctree\n");
      
      selection = GTK_CLIST(widget)->selection;
      if (!selection) 
	return;
      
      printf("selection\n");
      
      temp = gtk_ctree_node_get_row_data(GTK_CTREE(widget), selection->data);
      if (!temp)
	return;
      
      printf("temp\n");
      
      download = search_dup(temp, N_DOWNLOAD);
      handle_download(download, DOWNLOAD_NORMAL);*/
   } else if (event->button == 3) {
      temp_menu = gnome_popup_menu_new(data);
      gnome_popup_menu_do_popup_modal(temp_menu, NULL, NULL, event, widget);
      gtk_widget_destroy(temp_menu);
   }
}

void console_activate_cb() {
   char *console_text = NULL, *command, *message;
   
   console_text = gtk_entry_get_text(GTK_ENTRY(console_tab.entry));
   if (!console_text || !conn_info.connected) return;
   if (!strlen(console_text)) return;
   
   if (loc_end > MAX_HISTORY) {
      GList *link;
      
      link = g_list_nth(console_history, 0);
      console_history = g_list_remove_link(console_history, link);
      g_free(link->data);
      loc_end--;
   }
   console_history = g_list_append(console_history, strdup(console_text));
   loc_end++;
   loc_curr = loc_end;

   if (console_text[0] != '/') {
      do_sendtext(NULL, console_text);
      gtk_entry_set_text(GTK_ENTRY(console_tab.entry), "");
      return;
   }
   
   command = next_arg(console_text, &console_text);
   if (!command) {
      gtk_entry_set_text(GTK_ENTRY(console_tab.entry), "");
      return;
   }
   message = last_arg(console_text, &console_text);

   if (!handle_command(command, message)) {
      gnapster_text_insert(CURR, SYSTEM,
			   "Unknown command (try \2/help\1)\n");
   }
   
   gtk_entry_set_text(GTK_ENTRY(console_tab.entry), "");
}

int console_keypress_cb(GtkWidget *w, GdkEventKey *key) {
   GList *ptr;

   if (key->keyval >= GDK_1 && key->keyval <= GDK_9 && 
       key->state & GDK_MOD1_MASK) {
      gtk_notebook_set_page(GTK_NOTEBOOK(console_tab.notebook),
			    (key->keyval - 49));
      return 1;
   }
   
   if (key->keyval != GDK_Up && key->keyval != GDK_Down)
     return 0;
   
   switch(key->keyval) {
    case GDK_Up:
      loc_curr--;
      if (loc_curr < 0) loc_curr = 0;
      break;
    case GDK_Down:
      loc_curr++;
      if (loc_curr > loc_end) loc_curr = loc_end;
      break;
   }
   ptr = g_list_nth(console_history, loc_curr);
   if (ptr) {
      char *text;
      
      text = ptr->data;
      if (text)
	gtk_entry_set_text(GTK_ENTRY(w), text);
   } else if (loc_curr == loc_end) {
      gtk_entry_set_text(GTK_ENTRY(w), "");
   }
	
   gtk_signal_emit_stop_by_name(GTK_OBJECT(w), "key_press_event");
   
   return 1;
}

NAP_SEARCH *get_data(const void *row) {
   return ((GtkCListRow *)row)->data;
}

NAP_USER *get_chan_data(const void *row) {
   return ((GtkCListRow *)row)->data;
}

int compare_chan_user(GtkCList *clist, const void *row1, const void *row2) {
   NAP_USER *s1 = get_chan_data(row1), *s2 = get_chan_data(row2);
   if (!s1 || !s2)
     return 0;
   
   return g_strcasecmp(s1->user, s2->user);
}

int compare_chan_conn(GtkCList *clist, const void *row1, const void *row2) {
   NAP_USER *s1 = get_chan_data(row1), *s2 = get_chan_data(row2);
   if (!s1 || !s2)
     return 0;
   
   if (s1->conn <= s2->conn) return 1;
   if (s1->conn > s2->conn) return -1;
   
   return 0;
}

int compare_chan_shares(GtkCList *clist, const void *row1, const void *row2) {
   NAP_USER *s1 = get_chan_data(row1), *s2 = get_chan_data(row2);
   if (!s1 || !s2) 
     return 0;
   
   if (s1->shares <= s2->shares) return 1;
   if (s1->shares > s2->shares) return -1;
   
   return 0;
}

void channel_sort(GtkCList *clist, int column) {
   void *sortfunc[3] = {
      compare_chan_user, compare_chan_shares, compare_chan_conn
   };
   
   if (column >= 3 || column < 0)
     return;
   
   gtk_clist_set_compare_func(clist, sortfunc[column]);
   
   if (clist->sort_column == column) {
      clist->sort_type = (clist->sort_type == GTK_SORT_ASCENDING) ? 
			 GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
   }
   
   gtk_clist_set_sort_column(clist, column);
   gtk_clist_sort(clist);
}

int compare_file(GtkCList *clist, const void *row1, const void *row2) {
   NAP_SEARCH *s1 = get_data(row1);
   NAP_SEARCH *s2 = get_data(row2);
   
   return g_strcasecmp(s1->trunc_file, s2->trunc_file);
}

int compare_user(GtkCList *clist, const void *row1, const void *row2) {
   NAP_SEARCH *s1 = get_data(row1);
   NAP_SEARCH *s2 = get_data(row2);
   
   return g_strcasecmp(s1->user, s2->user);
}

int compare_size(GtkCList *clist, const void *row1, const void *row2) {
   NAP_SEARCH *s1 = get_data(row1);
   NAP_SEARCH *s2 = get_data(row2);
   
   if (s1->size <= s2->size) return 1;
   if (s1->size > s2->size) return -1;
   
   return 1;
}

int compare_length(GtkCList *clist, const void *row1, const void *row2) {
   NAP_SEARCH *s1 = get_data(row1);
   NAP_SEARCH *s2 = get_data(row2);
   int seconds1, seconds2;
   
   seconds1 = (s1->size / ((s1->kbps * 1000) / 8));
   seconds2 = (s2->size / ((s2->kbps * 1000) / 8));
   if (seconds1 <= seconds2) return 1;
   if (seconds1 > seconds2) return -1;
   
   return 1;
}

int compare_bitrate(GtkCList *clist, const void *row1, const void *row2) {
   NAP_SEARCH *s1 = get_data(row1);
   NAP_SEARCH *s2 = get_data(row2);
   
   if (s1->kbps <= s2->kbps) return 1;
   if (s1->kbps > s2->kbps) return -1;
   
   return 0;
}

int compare_connection(GtkCList *clist, const void *row1, const void *row2) {
   NAP_SEARCH *s1 = get_data(row1);
   NAP_SEARCH *s2 = get_data(row2);
   
   if (s1->connection <= s2->connection) return 1;
   if (s1->connection > s2->connection) return -1;
   
   return 0;
}

void search_sort(GtkCList *clist, int column) {
   void *sortfunc[6] = {
      compare_file, compare_connection, compare_user, compare_size, 
      compare_length, compare_bitrate
   };
   
   if (column >= 6 || column < 0)
     return;

   gtk_clist_set_compare_func(clist, sortfunc[column]);

   if (clist->sort_column == column) {
      clist->sort_type = (clist->sort_type == GTK_SORT_ASCENDING) ?
		    	 GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
   }
   
   gtk_clist_set_sort_column(clist, column);
   
   gtk_clist_sort(clist);
}

void search_activate_cb() {
   char *string, *media;
   char extra_optns[1024], temp[1024];
   
   if (!conn_info.connected || conn_info.searching) return;
   
   string = gtk_entry_get_text(GTK_ENTRY(search_tab.gtk_entry));
   if (!string || string[0] == 0) 
     return;
   string = strdup(string);

   gnapster_update_status(_("Searching..."), 1);
   conn_info.searching = 1;
   extra_optns[0] = '\0';
   if (user_info.conf[LINE_SPEED_SWITCH]) {
      char *cmp, *ptr;
      
      cmp = strdup(listint_to_str(user_info.conf[LINE_SPEED_CMP], 0));
      ptr = cmp;
      while(*ptr) {
	 *ptr = toupper(*ptr);
	 ptr++;
      }
      
      g_snprintf(temp, sizeof(temp), " LINESPEED \"%s\" %i", cmp, user_info.conf[LINE_SPEED]);
      strcat(extra_optns, temp);
      g_free(cmp);
   }
   if (user_info.conf[BITRATE_SWITCH]) {
      char *cmp, *ptr, *bit;
      
      cmp = strdup(listint_to_str(user_info.conf[BITRATE_CMP], 0));
      ptr = cmp;
      while(*ptr) {
	 *ptr = toupper(*ptr);
	 ptr++;
      }
      
      switch(user_info.conf[BITRATE]) {
       case 0:
	 bit = "96";
	 break;
       case 1:
	 bit = "112";
	 break;
       case 2:
	 bit = "128";
	 break;
       case 3:
	 bit = "160";
	 break;
       case 4:
	 bit = "192";
	 break;
       default:
	 bit = "";
	 break;
      }
      g_snprintf(temp, sizeof(temp), " BITRATE \"%s\" \"%s\"", cmp, bit);
      strcat(extra_optns, temp);
      g_free(cmp);
   }
   media = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(search_tab.media)->entry));
   if (g_strcasecmp(media, "audio") != 0) {
      g_snprintf(temp, sizeof(temp), " TYPE %s", media);
      strcat(extra_optns, temp);
   }
   g_strdown(string);
   napster_send(conn_info.sock, NAPSTER_SEARCH, "FILENAME CONTAINS \"%s\" MAX_RESULTS 100%s", 
		string, extra_optns);
   
   g_free(string);

   gtk_clist_clear(GTK_CLIST(search_tab.clist));
   gtk_clist_freeze(GTK_CLIST(search_tab.clist));
   gtk_entry_set_editable(GTK_ENTRY(search_tab.gtk_entry), 0);
}

void browse_user_menu(GtkWidget *clist) {
   void *temp;
   TYPE *n;
   int row = 0;
   char user[1024];
   
   if (!conn_info.connected) return;

   temp = gnapster_get_selection_data(clist, &row);
   if (!temp) return;
   
   n = (TYPE *)temp;
   
   if (n->type == N_SEARCH )
     g_snprintf(user, sizeof(user), ((NAP_SEARCH *)temp)->user);
   else if (n->type == N_DOWNLOAD)
     g_snprintf(user, sizeof(user), ((NAP_DOWNLOAD *)temp)->user);
   else if (n->type == N_USER)
     g_snprintf(user, sizeof(user), ((NAP_USER *)temp)->user);
   else 
     return;
   
   gnapster_update_status(_("Retrieving mp3 list..."), 1);
   conn_info.searching = 1;
   napster_send(conn_info.sock, NAPSTER_BROWSE, user);
   
   gtk_clist_clear(GTK_CLIST(browse_tab.ctree));
   gtk_clist_freeze(GTK_CLIST(browse_tab.ctree));
   gtk_entry_set_editable(GTK_ENTRY(browse_tab.entry), 0);
}

void browse_user_menu_cb(GtkWidget *w, gpointer data) {
   GtkWidget *data_widget;

   if (!data) return;

   data_widget = (GtkWidget *)data;
   browse_user_menu(data_widget);
}

void browse_user_cb() {
   char *string;
   
   if (GTK_CLIST(browse_tab.ctree)->freeze_count || !conn_info.connected ||
       conn_info.searching) 
     return;
   
   string = gtk_entry_get_text(GTK_ENTRY(browse_tab.entry));
   if (!strlen(string)) 
     return;
   
   gnapster_update_status(_("Retrieving mp3 list..."), 1);
   conn_info.searching = 1;
   napster_send(conn_info.sock, NAPSTER_BROWSE, string);
   
   gtk_clist_clear(GTK_CLIST(browse_tab.ctree));
   gtk_clist_freeze(GTK_CLIST(browse_tab.ctree));
   gtk_entry_set_editable(GTK_ENTRY(browse_tab.entry), 0);
}

void whois_user_menu_cb(GtkWidget *w, gpointer data) {
   GtkWidget *clist;
   int row = 0;
   NAP_SEARCH *temp;
   TYPE *n;
   char *user;
   
   if (!data) return;
   clist = (GtkWidget *)data;
   if (!clist) return;
   
   temp = gnapster_get_selection_data(clist, &row);
   if (!temp) return;
   
   n = (TYPE *)temp;
   
   if (n->type == N_SEARCH)
     user = temp->user;
   else if (n->type == N_USER) 
     user = ((NAP_USER *)temp)->user;
   else 
     return;

   gnapster_text_insert(CURR, SYSTEM, 
			"Sending whois request for \2%s\1...\n", user);
   napster_send(conn_info.sock, NAPSTER_WHOIS_REQUEST, user);
}

void connect_cb() {
   if (conn_info.connected || conn_info.connecting) return;
   if (!user_info.user) {
      gnapster_update_status(_("Please edit your preferences before attempting to connect!"), 0);
      return;
   }
   
   close(conn_info.sock);
   if (conn_info.sock_input) gdk_input_remove(conn_info.sock_input);
   conn_info.connected = 0;
   conn_info.connecting = 1;
   
   g_free(user_info.server);
   user_info.server = gnome_config_get_string("/gnapster/Options/server");
   user_info.server_port = gnome_config_get_int("/gnapster/Options/server_port");
   
   /* Napster servers changed their IP...this compensates... */
   if (!user_info.server || !strcmp(user_info.server, "208.184.216.223")) {
      gnome_config_set_string("/gnapster/Options/server", "208.184.216.222");
      gnome_config_set_int("/gnapster/Options/server_port", 8875);
      gnome_config_sync();
      user_info.server = gnome_config_get_string("/gnapster/Options/server");
      user_info.server_port = gnome_config_get_int("/gnapster/Options/server_port");
   }
   if (!user_info.server_port) {
      gnome_config_set_int("/gnapster/Options/server_port", 8875);
      gnome_config_sync();
      user_info.server_port = gnome_config_get_int("/gnapster/Options/server_port");
   }
   
   if (user_info.server_port == 8875) {
      /* Query server... */
      gnapster_update_status(_("Finding best host..."), 1);
      napster_get_best_host();
   } else {
      g_free(conn_info.ip);
      conn_info.ip = strdup(user_info.server);
      conn_info.port = user_info.server_port;
      conn_info.connecting = 1;
      conn_info.connected = 0;

      gnapster_update_status(_("Connecting to server..."), 1);
      napster_connect(conn_info.ip, conn_info.port);
   }
}

void disconnect_cb() {
   GList *ptr;
   
   if (!conn_info.connected && !conn_info.connecting) return;
   
   close(conn_info.sock);
   if (conn_info.sock_input) 
     gdk_input_remove(conn_info.sock_input);
   if (conn_info.upload_input) 
     gdk_input_remove(conn_info.upload_input);
   conn_info.connected = 0;
   conn_info.connecting = 0;
   conn_info.searching = 0;
   
   gtk_label_set_text(GTK_LABEL(upload_tab.label), "");
   gtk_entry_set_editable(GTK_ENTRY(search_tab.gtk_entry), 1);
   gtk_widget_set_sensitive(search_tab.media, 0);
   
   gnapster_update_status(_("Disconnected"), 0);
   gtk_text_set_point(GTK_TEXT(motd_tab.text), 0);
   gtk_text_forward_delete(GTK_TEXT(motd_tab.text), 
			   gtk_text_get_length(GTK_TEXT(motd_tab.text)));
   
   ptr = console_tab.channel_list;
   while(ptr) {
      if (((CHANNEL *)ptr->data)->clist)
	gtk_clist_clear(GTK_CLIST(((CHANNEL *)ptr->data)->clist));
      gtk_notebook_remove_page(GTK_NOTEBOOK(console_tab.notebook),
			       1);
      g_free(((CHANNEL *)ptr->data)->channel_name);
      g_free(ptr->data);
      ptr = ptr->next;
   }
   g_list_free(console_tab.channel_list);
   console_tab.channel_list = NULL;
}

void search_download_cb(GtkWidget *w, gpointer data) {
   download_true_cb(search_tab.clist, data);
}

void browse_download_cb(GtkWidget *w, gpointer data) {
   download_true_cb(browse_tab.ctree, data);
}

void cancel_upload_cb() {
   NAP_DOWNLOAD *temp;
   GtkWidget *clist = (GtkWidget *)upload_tab.clist;
   
   temp = gnapster_get_selection_data(clist, NULL);
   if (!temp) return;
   
   upload_close(temp, TRANSFER_CANCEL);
}

char *listint_to_str(int i, int type) {
   char *cmp[] = { 
      "At least", "Equal to", "At best", NULL
   }, *conn[] = {
      "Unknown", "14.4", "28.8", "33.6", "56k", "64k ISDN", "128k ISDN", "Cable", "DSL",
      "T1", "T3+", NULL
   };
   
   switch(type) {
    case 0:
      return cmp[i];
    case 1:
      return conn[i];
    default:
      return NULL;
   }
}

int liststr_to_int(char *s) {
   char *conn[] = {
      "Unknown", "14.4", "28.8", "33.6", "56k", "64k ISDN", "128k ISDN", "Cable", "DSL",
      "T1", "T3+", NULL
   };
   char *bitrate[] = {
      "96", "112", "128", "160", "192", NULL
   };
   int i = 0;

   if (!strcmp(s, "At least"))
     return 0;
   else if (!strcmp(s, "Equal to"))
     return 1;
   else if (!strcmp(s, "At best"))
     return 2;
   
   while(conn[i]) {
      if (!strcmp(s, conn[i])) return i;
      i++;
   }
   i = 0;
   while(bitrate[i]) {
      if (!strcmp(s, bitrate[i])) return i;
      i++;
   }

   return 0;
}

void combo_changed(GtkWidget *widget, gpointer data) {
   WidgetProperty *prop;
   int val;
   
   prop = gtk_object_get_data(GTK_OBJECT(widget), "prop");
   if (!prop)
     return;
   
   if (GTK_IS_SPIN_BUTTON(widget))
     val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
   else
     val = liststr_to_int(gtk_entry_get_text(GTK_ENTRY(widget)));
   user_info.conf[prop->key] = val;

   gnome_config_set_int(prop->conf, val);
   gnome_config_sync();
}

void preferences_changed_cb() {
   if (!prop_dlg.window) return;
   gnome_property_box_changed(GNOME_PROPERTY_BOX(prop_dlg.window));
}

void preferences_button_cb(GtkWidget *widget, gint arg1, gpointer data) {
   char *user, *pass, *nap_port, *conn, *download_dir, *upload_dir;

   prop_dlg.calling++;
   if (prop_dlg.calling == 2) {
      prop_dlg.calling = 0;
      return;
   }
   
   user = gtk_entry_get_text(GTK_ENTRY(prop_dlg.user_entry));
   pass = gtk_entry_get_text(GTK_ENTRY(prop_dlg.pass_entry));
   nap_port = gtk_entry_get_text(GTK_ENTRY(prop_dlg.nap_port));
   conn = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(prop_dlg.connection)->entry));
   download_dir = gtk_entry_get_text(GTK_ENTRY(gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(prop_dlg.download_dir))));
   upload_dir = gtk_entry_get_text(GTK_ENTRY(gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(prop_dlg.upload_dir))));
   
//   server = gtk_entry_get_text(GTK_ENTRY(gnome_entry_gtk_entry(GNOME_ENTRY(prop_dlg.server_entry))));

   if (!strlen(user) || !strlen(pass) || !strlen(conn) || !strlen(download_dir) || !strlen(nap_port)) {
      gnapster_update_status(_("Form not complete!  Make sure you fill in everything!"), 0);
      return;
   }
   
   user_info.connection = napster_conn_as_int(conn);
   if (my_atoi(nap_port) != conn_info.nap_port || user_info.conf[FIREWALL] == 0) {
      conn_info.nap_port = my_atoi(nap_port);
      gdk_input_remove(conn_info.bind_sock_input);
      close(conn_info.bind_sock);
      if (conn_info.nap_port)
	napster_bind(conn_info.nap_port);
      if (conn_info.connected) {
	 napster_send(conn_info.sock, NAPSTER_CHANGE_PORT, "%i",
		      conn_info.nap_port);
      }
      gnome_config_set_int("/gnapster/Options/port", conn_info.nap_port);
   } else if (user_info.conf[FIREWALL] == 1) {
      conn_info.nap_port = 0;
      gdk_input_remove(conn_info.bind_sock_input);
      close(conn_info.bind_sock);
      if (conn_info.connected)
	napster_send(conn_info.sock, NAPSTER_CHANGE_PORT, "0");
   }
   
   gnome_config_set_string("/gnapster/User/user", user);
   gnome_config_set_string("/gnapster/User/pass", pass);
   gnome_config_set_int("/gnapster/User/connection", user_info.connection);
   gnome_config_set_string("/gnapster/User/dl_dir", download_dir);
   if (strlen(upload_dir)) 
     gnome_config_set_string("/gnapster/User/ul_dir", upload_dir);
   
   gnapster_config_set_int(0);
   
   g_free(user_info.user);
   g_free(user_info.pass);
   g_free(user_info.download_dir);
   g_free(user_info.upload_dir);
   user_info.user = gnome_config_get_string("/gnapster/User/user");
   user_info.pass = gnome_config_get_string("/gnapster/User/pass");
   user_info.download_dir = gnome_config_get_string("/gnapster/User/dl_dir");   
   user_info.upload_dir = gnome_config_get_string("/gnapster/User/ul_dir");
   
   if (!conn_info.connected && !conn_info.connecting)
     gnapster_update_status(_("Not connected (use the file menu to connect)"), -1);
}

void preferences_destroy_cb(GtkWidget *widget, gint arg1, gpointer data) {
   prop_dlg.window = NULL;
}

void clicked_cb(GtkWidget *widget, int *data) {
   WidgetProperty *prop;
   GList *ptr;
   
   *data = !*data;
   
   prop = gtk_object_get_data(GTK_OBJECT(widget), "prop");
   if (!prop)
     return;
   
   gnome_config_set_int(prop->conf, *data);
   ptr = prop->sensitive;
   while(ptr) {
      gtk_widget_set_sensitive(GTK_WIDGET(ptr->data), *data);
      ptr = ptr->next;
   }
   
   gnome_config_sync();
}

void preferences_clicked_cb(GtkWidget *widget, int *data) {
   clicked_cb(widget, data);
   preferences_changed_cb();
}

void preferences_cb() {
   if (prop_dlg.window) return;
   
   create_prop_dlg(preferences_changed_cb,
		   preferences_button_cb,
		   preferences_destroy_cb,
		   preferences_clicked_cb,
		   &(user_info.new_user));
   gtk_widget_show_all(prop_dlg.window);
}

void build_dlg_cancel_cb() {
   building = 0;
   gnome_dialog_close(GNOME_DIALOG(build_dlg.window));
   gdk_input_remove(build_dlg.pipe_input);
   if (build_dlg.cpid)
     kill(build_dlg.cpid, SIGUSR1);
}

void about_cb() {
   GtkWidget *about_box;
   const char *authors[] = { "jasta <jasta@gotlinux.org>", NULL };
   
   about_box = gnome_about_new(_("Gnapster"), VERSION,
			       _("Copyright (C) 1999-2000 Josh Guilfoyle"),
			       authors,
			       _("Napster (online mp3 community) client written for GNOME"),
			       NULL);
   gtk_widget_show(about_box);
}
