#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>

#include <gnome.h>

#include "napster.h"
#include "gnapster.h"
#include "callbacks.h"
#include "ui.h"
#include "upload.h"
#include "network.h"
#include "commands.h"
#include "download.h"
#include "queue.h"

#include "chan.h"

#include "config.h"

extern LIST_CHAN_DLG list_chan_dlg;
extern CONN_INFO conn_info;
extern GNAPSTER_MAIN gnapster_main;
extern CONSOLE_TAB console_tab;
extern PROP_DLG prop_dlg;
extern USER_INFO user_info;

void nick_completion_fill_nicks(GCompletion *comp, GtkCList *clist) {
   NAP_USER *dptr;
   GList *nick_list = NULL, *ptr;
   
   for(ptr=clist->row_list; ptr; ptr=ptr->next) {
      dptr = GTK_CLIST_ROW(ptr)->data;
      if (!dptr)
	break;
      
      if (dptr->user && strcmp(dptr->user, user_info.user))
	nick_list = g_list_append(nick_list, all_lower(dptr->user));
   }
   
   g_completion_add_items(comp, nick_list);
}

char *nick_comp_func(void *data) {
   GtkCListRow *row;
   NAP_USER *nuser;
   
   row = data;
   nuser = row->data;
   
   return g_strdup(all_lower(nuser->user));
}

char *nick_completion(char *ptr, GtkCList *clist) {
   char *nick, *cnick, *text, *new_buf, *npfx;
   GCompletion *comp;
   GList *nicks;
   
   if (!ptr)
     return NULL;
   
   new_buf = strdup(ptr);
   
   if (!strchr(ptr, ':'))
     return new_buf;
   
   nick = next_arg_full(ptr, &ptr, ':');
   text = ptr;
   
   /* Not the first word */
   if (nick && (strchr(nick, ' ') || strlen(nick) <= 1))
     return new_buf;
   
   comp = g_completion_new(nick_comp_func);
   g_completion_add_items(comp, clist->row_list);
//   nick_completion_fill_nicks(comp, clist);
   
   nicks = g_completion_complete(comp, nick, &npfx);
   if (!nicks || !npfx) {
      g_completion_free(comp);
      return new_buf;
   }
   
   cnick = ((NAP_USER *)((GtkCListRow *)g_list_nth_data(nicks, 0))->data)->user;
   if (!strcmp(cnick, user_info.user))
     cnick = ((NAP_USER *)((GtkCListRow *)g_list_nth_data(nicks, 0))->data)->user;
   
   if (!cnick) {
      g_free(npfx);
      g_completion_free(comp);
      return new_buf;
   }
   
   g_free(new_buf);
   new_buf = g_strdup_printf("%s:%s", cnick, text);
   
   g_free(npfx);
   g_completion_free(comp);
   
   return new_buf;
}

void autojoin_handle() {
   FILE *f;
   char *chan, *dptr;
   GList *ptr, *felem;
   
   f = open_autojoin("r");
   if (!f)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   if (felem)
     gnapster_text_insert(CONSOLE, SYSTEM, "Auto-joining channels...\n");
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      chan = next_arg(dptr, &dptr);
      if (chan)
	napster_send(conn_info.sock, NAPSTER_JOIN_CHAN, chan);
      
      g_free(chan);
   }
   
   g_list_free(felem);
}

FILE *open_autojoin(char *flags) {
   FILE *f;
   char *conf;
   
   conf = g_strdup_printf("%s/.gnapster/autojoin", getenv("HOME"));
   
   f = fopen(conf, flags);
   
   if (!f)
     perror("fopen");
   
   g_free(conf);
   
   return f;
}

void append_autojoin_entry(char *chan) {
   FILE *f;
   char *dptr;
   GList *ptr, *felem;
   
   if (!chan)
     return;
   
   f = open_autojoin("r");
   if (f) {
      felem = read_file(f);

      fclose(f);

      for(ptr=felem; ptr; ptr=ptr->next) {
	 dptr = ptr->data;

	 if (dptr && !strncmp(dptr, chan, strlen(chan)))
	   break;

	 g_free(dptr);
      }

      g_list_free(felem);
   }

   f = open_autojoin("a");
   if (!f)
     return;

   fprintf(f, "%s\n", chan);
   
   fclose(f);
}

void remove_autojoin_entry(char *chan) {
   FILE *f;
   char *dptr;
   GList *ptr, *felem;
   
   if (!chan)
     return;
   
   f = open_autojoin("r");
   if (!f)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   f = open_autojoin("w");
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      if (dptr && strncmp(dptr, chan, strlen(chan)))
	fprintf(f, "%s", dptr);

      g_free(dptr);
   }
   
   g_list_free(felem);
   
   fclose(f);
}

void autojoin_add() {
   char *chan;
   
   chan = gtk_entry_get_text(GTK_ENTRY(prop_dlg.aj_entry));
   if (!strlen(chan))
     return;
   
   append_autojoin_entry(chan);
   
   gtk_entry_set_text(GTK_ENTRY(prop_dlg.aj_entry), "");
   
   populate_autojoin_clist();
}

void autojoin_remove() {
   char *chan;
   int row = -1;
   
   gnapster_get_selection_data(prop_dlg.aj_clist, &row);
   if (row == -1)
     return;
   
   gtk_clist_get_text(GTK_CLIST(prop_dlg.aj_clist), row, 0, &chan);
   
   remove_autojoin_entry(chan);
   
   populate_autojoin_clist();
}

void populate_autojoin_clist() {
   FILE *f;
   char *dptr;
   GList *ptr, *felem;
   
   if (!prop_dlg.aj_clist)
     return;
   
   f = open_autojoin("r");
   if (!f)
     return;

   if (prop_dlg.aj_clist) {
      gtk_clist_freeze(GTK_CLIST(prop_dlg.aj_clist));
      gtk_clist_clear(GTK_CLIST(prop_dlg.aj_clist));
   }
   
   felem = read_file(f);
   
   fclose(f);
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      gnapster_clist_append(prop_dlg.aj_clist, NULL, NULL, dptr, NULL);
      
      g_free(dptr);
   }
   
   g_list_free(felem);
   
   gtk_clist_thaw(GTK_CLIST(prop_dlg.aj_clist));
}

void list_chan_dlg_clist_cb(GtkWidget *widget, GdkEventButton *event, gpointer data) {
   int row = -1, col;
   
   gtk_clist_get_selection_info(GTK_CLIST(widget), event->x, event->y,
				&row, &col);
   
   if (event->type == GDK_2BUTTON_PRESS && event->button == 1)
     list_chan_dlg_cb(NULL, "join");
}

void list_chan_dlg_cb(GtkWidget *w, void *data) {
   if (!data)
     return;
   
   if (!strcmp(data, "join")) {
      NAP_CHANNEL *nchan;
      
      nchan = gnapster_get_selection_data(list_chan_dlg.clist, NULL);
      if (!nchan)
	return;
      
      napster_send(conn_info.sock, NAPSTER_JOIN_CHAN, nchan->chan);
   }
   
   gnome_dialog_close(GNOME_DIALOG(list_chan_dlg.window));
   list_chan_dlg.window = NULL;
}

void list_chan_cb(GtkWidget *w, gpointer data) {
   if (!conn_info.connected || conn_info.connecting)
     return;
   
   create_list_chan_dlg();
   
   gtk_clist_freeze(GTK_CLIST(list_chan_dlg.clist));
   
   napster_send(conn_info.sock, NAPSTER_CHAN_LIST, NULL);
}

void join_chan_cb(GtkWidget *w, gpointer data) {
   gtk_notebook_set_page(GTK_NOTEBOOK(gnapster_main.notebook), 4);
   gtk_entry_set_text(GTK_ENTRY(console_tab.entry), "/join");
   gtk_widget_grab_focus(console_tab.entry);
}

void part_chan_cb(GtkWidget *w, gpointer data) {
   if (!remove_console_tab(0))
     gnome_error_dialog("You must click on a channel tab before attempting to remove it");
}

int chan_lookup(char *chan) {
   GList *ptr;
   int i = 0;
   
   for(ptr=console_tab.channel_list; ptr; ptr=ptr->next) {
      if (!g_strcasecmp(((CHANNEL *)ptr->data)->channel_name, chan) &&
	  !((CHANNEL *)ptr->data)->user)
	return (i + 2);
      
      i++;
   }
   
   return -1;
}

GtkCList *chan_lookup_clist(char *chan) {
   GList *ptr;
   
   for(ptr=console_tab.channel_list; ptr; ptr=ptr->next) {
      if (!g_strcasecmp(((CHANNEL *)ptr->data)->channel_name, chan) &&
	  !((CHANNEL *)ptr->data)->user)
	return GTK_CLIST(((CHANNEL *)ptr->data)->clist);
   }
   
   return NULL;
}

char *chan_lookup_chan(int index) {
   CHANNEL *chan;
   
   chan = g_list_nth_data(console_tab.channel_list, index);
   if (!chan)
     return NULL;
   
   if (!chan->user)
     return chan->channel_name;
   
   return NULL;
}

