/* gnapster.c - controls main loop and various run-time GUI changes */

#include <stdio.h>

#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>

#include <signal.h>

#include <gnome.h>

#include <sys/utsname.h>
#include <ctype.h>

#include "napster.h"
#include "gnapster.h"
#include "ui.h"
#include "callbacks.h"

#include "upload.h"
#include "network.h"
#include "chan.h"
#include "resume.h"

#include "config.h"
#include "timestamp.h"

#include "../pixmaps/speedgreen.xpm"
#include "../pixmaps/speedyellow.xpm"
#include "../pixmaps/speedred.xpm"
#include "../pixmaps/speedgray.xpm"

GNAPSTER_MAIN gnapster_main;
SEARCH_TAB search_tab;
BROWSE_TAB browse_tab;
DOWNLOAD_TAB download_tab;
UPLOAD_TAB upload_tab;
CONSOLE_TAB console_tab;
MOTD_TAB motd_tab;

PROP_DLG prop_dlg;
BUILD_DLG build_dlg;
BROWSE_DLG browse_dlg;
ADD_DLG add_dlg;
EDIT_DLG edit_dlg;
LIST_CHAN_DLG list_chan_dlg;
RESUME_DLG resume_dlg;
REFRESH_DLG refresh_dlg;

CONN_INFO conn_info;
USER_INFO user_info;

int building;
GList *total_list = NULL;
GList *ignore_list = NULL;
GList *signore_list = NULL;

GdkPixmap *speedgreen = NULL, *speedyellow = NULL, *speedred = NULL, *speedgray = NULL;
GdkBitmap *speedgreenb = NULL, *speedyellowb = NULL, *speedredb = NULL, *speedgrayb = NULL;

CONFIG_TABLE conf_table[] = {
  /* Key                        Gnapster key		Default Value */
   { "name_fix", 		NAME_FIX,		1 },
   { "theme_colors", 		THEME_COLORS,		0 },
   { "auto_rm", 		AUTO_RM,		0 },
   { "opennap_server", 		OPENNAP_SERVER,		0 },
   { "auto_start_queue", 	AUTO_START_QUEUE,	1 },
   { "line_speed_switch", 	LINE_SPEED_SWITCH,	0 },
   { "line_speed_cmp", 		LINE_SPEED_CMP,		0 },
   { "line_speed", 		LINE_SPEED,		0 },
   { "bitrate_switch", 		BITRATE_SWITCH,		0 },
   { "bitrate_cmp", 		BITRATE_CMP,		0 },
   { "bitrate", 		BITRATE,		0 },
   { "max_uploads_switch", 	MAX_UPLOADS_SWITCH,	1 },
   { "max_uploads", 		MAX_UPLOADS,		5 },
   { "reject_ul", 		REJECT_UL,		0 },
   { "firewall", 		FIREWALL,		0 },
   { "max_peruser_uploads_switch", MAX_PERUSER_UPLOADS_SWITCH,	0 },
   { "max_peruser_uploads",	MAX_PERUSER_UPLOADS,	2 },
   { "autoquery",		AUTOQUERY,		1 },
   { "max_downloads_switch",	MAX_DOWNLOADS_SWITCH,	0 },
   { "max_downloads",		MAX_DOWNLOADS,		4 },
   { NULL, 		       -1,		       -1 },
};

FREE_TABLE gnapster_free[] = {
   { N_SEARCH,		search_free },
   { N_DOWNLOAD,	download_free },
   { N_SERVER,		server_free },
   { N_USER,		user_free },
   { N_CHANNEL,		channel_free },
   { N_RESUME,		resume_free },
   { W_PROP,		property_free },
   { -1,		NULL },
};

GList *read_file(FILE *f) {
   char buf[1024];
   GList *list = NULL;
   
   while(!feof(f)) {
      fgets(buf, sizeof(buf) - 1, f);
      if (feof(f)) break;
      if (buf[0] && buf[0] != '\n')
	list = g_list_append(list, strdup(buf));
   }
   
   return list;
}

NAP_DOWNLOAD *gnapster_match_real_download(char *user, char *file, int *row) {
   NAP_DOWNLOAD *ret;
   GList *ptr;
   int i = 0;
   
   ptr = GTK_CLIST(download_tab.clist)->row_list;
   if (!ptr)
     return NULL;
   while(ptr) {
      ret = gtk_clist_get_row_data(GTK_CLIST(download_tab.clist), i);
      if (!strcmp(ret->file, file) &&
	  !strcmp(ret->user, user)) {
	 if (row)
	   *row = i;
	 return ret;
      }
      ptr = ptr->next, i++;
   }
   
   return NULL;
}

/* Unused function */
void *gnapster_new(size_t v, int t) {
   void *ret;
   
   ret = g_malloc(sizeof(v));
   if (!ret) return NULL;
   
   memset(ret, 0, sizeof(v));
   ((TYPE *)ret)->type = t;
   
   return ret;
}

int gnapster_get_type(void *data) {
   int t = -1;
   
   if (data)
     t = ((TYPE *)data)->type;
   
   return t;
}

char *gnapster_sv() {
   static char ver[256];
   struct utsname uts;
   
   if (uname(&uts) < 0)
     return "Gnapster 1.3.8";
   
   g_snprintf(ver, sizeof(ver), "%s [ %s ] / %s %s %s",
	      NAP_VERSION, TIMESTAMP, uts.nodename, uts.sysname, uts.release);
   
   return ver;
}

int gnapster_key_lookup(char *s) {
   int i = 0;
   
   if (!s) return -1;
   
   while(conf_table[i].conf_str) {
      if (!strcmp(conf_table[i].conf_str, s))
	return i;
      i++;
   }
   
   return -1;
}

void gnapster_config_combo_widget(GtkWidget *w, char *key) {
   WidgetProperty *prop;
   char *full_conf;
   GtkWidget *widget;
   int conf;
   
   if (GTK_IS_SPIN_BUTTON(w))
     widget = w;
   else
     widget = GTK_COMBO(w)->entry;
   
   conf = gnapster_key_lookup(key);
   if (conf < 0)
     fprintf(stderr, "Warning, key '%s' not found!\n", key);

   prop = g_malloc(sizeof(WidgetProperty));
   memset(prop, 0, sizeof(WidgetProperty));
   full_conf = g_strdup_printf("/gnapster/Options/%s", key);
   prop->key = conf;
   prop->conf = full_conf;
   gtk_object_set_data_full(GTK_OBJECT(widget), "prop", prop, gnapster_destroy);
   gtk_signal_connect(GTK_OBJECT(widget), "changed",
		      GTK_SIGNAL_FUNC(combo_changed), NULL);
}

void gnapster_config_toggle_widget(GtkWidget *w, char *key, ...) {
   WidgetProperty *prop;
   GList *sensitive = NULL;
   char *full_conf;
   GtkWidget *widget;
   int conf;
   va_list args;
   
   conf = gnapster_key_lookup(key);
   if (conf < 0)
     fprintf(stderr, "Warning, key '%s' not found!\n", key);
   
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), user_info.conf[conf]);
   
   prop = g_malloc(sizeof(WidgetProperty));
   memset(prop, 0, sizeof(WidgetProperty));
   full_conf = g_strdup_printf("/gnapster/Options/%s", key);
   prop->key = conf;
   prop->conf = full_conf;
   va_start(args, key);
   while(1) {
      widget = va_arg(args, GtkWidget *);
      if (!widget) 
	break;
      sensitive = g_list_append(sensitive, widget);
   }
   va_end(args);
   prop->sensitive = sensitive;
   gtk_object_set_data_full(GTK_OBJECT(w), "prop", prop, gnapster_destroy);
   gtk_signal_connect(GTK_OBJECT(w), "clicked",
		      GTK_SIGNAL_FUNC(clicked_cb), &user_info.conf[conf]);
}

int upload_accepted(char *user, char *file, char *tf) {
   int uploads;
   
   uploads = g_list_length(GTK_CLIST(upload_tab.clist)->row_list);
   
   if (user_info.conf[MAX_UPLOADS_SWITCH] &&
       uploads >= user_info.conf[MAX_UPLOADS]) {
      gnapster_text_insert(CONSOLE, SYSTEM, "Rejecting upload request for \2%s\1 to \2%s\1, max uploads reached\n",
			   tf, user);
      napster_send(conn_info.sock, NAPSTER_SENDLIMIT,
		   "%s \"%s\" %i", user, file,
		   user_info.conf[MAX_UPLOADS]);
      
      return 0;
   }
   
   if (user_info.conf[REJECT_UL] &&
       g_list_length(GTK_CLIST(download_tab.clist)->row_list)) {
      gnapster_text_insert(CONSOLE, SYSTEM, "Rejecting upload request for \2%s\1 to \2%s\1, active downloads found\n",
			   tf, user);
      napster_send(conn_info.sock, NAPSTER_SENDLIMIT,
		   "%s \"%s\" 1", user, file);
      
      return 0;
   }
   
   if (user_info.conf[MAX_PERUSER_UPLOADS_SWITCH] &&
	      napster_user_uploads(user) >= user_info.conf[MAX_PERUSER_UPLOADS]) {
      gnapster_text_insert(CONSOLE, SYSTEM, "Rejecting upload request for \2%s\1 to \2%s\1, max per user uploads reached\n",
			   tf, user);
      napster_send(conn_info.sock, NAPSTER_SENDLIMIT,
		   "%s \"%s\" %i", user, file,
		   user_info.conf[MAX_PERUSER_UPLOADS]);
      
      return 0;
   }
   
   return 1;
}

int verify_file_shared(char *sfile) {
   FILE *f;
   char *mtime, *file;
   char *ptr, buf[1024 * 6];
   
   if (!sfile)
     return 0;
   
   f = open_shared(0);
   if (!f)
     return 0;
   
   while(!feof(f)) {
      fgets(buf, sizeof(buf) - 1, f);
      if (feof(f)) break;
      ptr = buf;
      mtime = next_arg(ptr, &ptr);
      file = next_arg(ptr, &ptr);
      if (!file)
	continue;
      if (!strcmp(file, sfile))
	return 1;
   }
   
   return 0;
}

char *all_lower(char *s) {
   int x = strlen(s);
   static char ret[1024 * 6];
   
   if (!s)
     return NULL;
 
   if (x > sizeof(ret)) 
     return s;
   
   memset(ret, 0, sizeof(ret));
   
   while(x--)
     ret[x] = tolower(s[x]);
   
   return ret;
}

void conv_file(char *s) {
   char *ptr;

   if (!user_info.conf[NAME_FIX]) return;
   
   ptr = strrchr(s, '/');
   if (!ptr)
     return;
   while(*ptr) {
      if (*ptr == ' ') *ptr = '_';
      ptr++;
   }
}

FILE *open_file(NAP_DOWNLOAD *download, char *file) {
   FILE *f;
   char *temp_str, *temp_str2 = NULL;
   int inc = 1;
   
   temp_str = g_strdup_printf("%s/%s", user_info.download_dir, file);
   conv_file(temp_str);
   if (file_exists(temp_str)) {
      temp_str2 = g_strdup_printf("%s.%i", temp_str, inc);
      conv_file(temp_str2);
      while(file_exists(temp_str2)) {
	 g_free(temp_str2);
	 temp_str2 = g_strdup_printf("%s.%i", temp_str, inc);
	 conv_file(temp_str2);
	 inc++;
      }
   } else 
     temp_str2 = strdup(temp_str);
   
   download->fpath = strdup(temp_str2);
   f = fopen(download->fpath, "w");
   
   if (strcmp(temp_str, temp_str2)) {
      char *temp_str3;
      temp_str3 = g_strdup_printf("File already exists, using backup file \2%s\1...\n",
				  temp_str2);
      gnapster_text_insert(CONSOLE, SYSTEM, temp_str3);
      g_free(temp_str3);
   }
   
   g_free(temp_str);
   g_free(temp_str2);
   
   return f;
}

int file_exists(char *file) {
   struct stat st;
   int x;
   
   x = stat(file, &st);
   if (x < 0) return 0;
   
   return 1;
}

/* Totally destroys an MP3 GList */
GList *mp3_list_free(GList *list) {
   GList *ptr;
   MP3 *mp3;
   
   if (!list) return NULL;

   ptr = list;
   while(ptr) {
      mp3 = (MP3 *)ptr->data;
      if (mp3) {
	 g_free(mp3->filename);
	 g_free(mp3->checksum);
	 g_free(mp3);
	 ptr->data = NULL;
      }
      list = g_list_remove_link(list, ptr);
      ptr = ptr->next;
   }
   
   return list;
}

char *next_arg_full(char *s, char **next, char end) {
   char *ptr, *str;
   
   /* Checks if the entire string is NULL or that we've pushed the
    * string to become the \0 */
   if (!s || !*s) return NULL;
   
   ptr = s;
   
   str = (end == '"') ? ++ptr : ptr;
//   if (*str == end) return " ";
   while(*str && *str != end && *str != '\n')
     str++;
   if (*str) {
      *str++ = 0;
      if (end == '"') str++;
   }
   
   if (next)
     *next = str;
   
   return ptr;
}

char *next_arg_first(char *s, char **next, char **first) {
   *first = NULL;
   
   if (!s || !*s)
     return NULL;
   
   *first = s;
   
   return next_arg(s, next);
}

char *next_arg(char *s, char **next) {
   char cmd;
   
   if (!s || !*s) return NULL;
   
   if (*s == '"') cmd = '"';
   else cmd = ' ';
   
   return next_arg_full(s, next, cmd);
}

char *last_arg(char *s, char **next) {
   return next_arg_full(s, next, 0);
}

void *download_dup(NAP_DOWNLOAD *download) {
   NAP_DOWNLOAD *temp;
   
   if (!download)
     return NULL;

   temp = g_malloc(sizeof(NAP_DOWNLOAD));
   memset(temp, 0, sizeof(NAP_DOWNLOAD));
   
   temp->type = N_DOWNLOAD;
   temp->trunc_file = strdup(download->trunc_file);
   temp->file = strdup(download->file);
   temp->user = strdup(download->user);
   temp->id = strdup(download->id);
   temp->num = download->num;
   temp->kbps = download->kbps;
   temp->freq = download->freq;
   temp->connection = download->connection;
   if (download->resume_path)
     temp->resume_path = strdup(download->resume_path);
   
   return temp;
}

void *search_dup(NAP_SEARCH *search, int type) {
   if (!search) 
     return NULL;
   
   if (type == N_SEARCH) {
      NAP_SEARCH *temp;
      temp = g_malloc(sizeof(NAP_SEARCH));
      memset(temp, 0, sizeof(NAP_SEARCH));
      temp->type = N_SEARCH;
      temp->trunc_file = strdup(search->trunc_file);
      temp->file = strdup(search->file);
      temp->user = strdup(search->user);
      temp->id = strdup(search->id);
      temp->num = search->num;
      temp->size = search->size;
      temp->kbps = search->kbps;
      temp->freq = search->freq;
      temp->connection = search->connection;
      temp->ip = search->ip;
      return temp;
   } else if (type == N_DOWNLOAD) {
      NAP_DOWNLOAD *temp;
      temp = g_malloc(sizeof(NAP_DOWNLOAD));
      memset(temp, 0, sizeof(NAP_DOWNLOAD));
      temp->type = N_DOWNLOAD;
      temp->trunc_file = strdup(search->trunc_file);
      temp->file = strdup(search->file);
      temp->user = strdup(search->user);
      temp->id = strdup(search->id);
      temp->num = search->num;
//      temp->size = search->size;
      temp->kbps = search->kbps;
      temp->freq = search->freq;
      temp->connection = search->connection;
//      temp->ip = search->ip;
      if (search->resume_path)
	temp->resume_path = strdup(search->resume_path);
      return temp;
   }
   return NULL;
}

long my_atol(const char *nptr) {
   if (!nptr) return -1;
   
   return atol(nptr);
}

int my_atoi(const char *nptr) {
   if (!nptr) return -1;
   
   return atoi(nptr);
}

NAP_DOWNLOAD *gnapster_parse_download(char *s, int alt) {
   char *trunc_file, *file, *user, *id, *ip_str;
   int port, connection;
   unsigned long int ip = 0;
   NAP_DOWNLOAD *temp;
   
   user = next_arg(s, &s);
//   ip = my_atol(next_arg(s, &s));
   ip_str = next_arg(s, &s);
   port = my_atoi(next_arg(s, &s));
   file = next_arg(s, &s);
   id = next_arg(s, &s);
   connection = my_atoi(next_arg(s, &s));
   if (connection < 0) return NULL;
   
   sscanf(ip_str, "%lu", &ip);
   
   trunc_file = strrchr(file, '\\');
   if (!trunc_file) return NULL;
   trunc_file++;

   if (port == 0 && user_info.conf[FIREWALL] && !alt) {
      NAP_DOWNLOAD *real;
      gnapster_text_insert(CONSOLE, SYSTEM, "Unable to receive from \2%s\1, user cannot accept connections\n",
			   user);
      real = gnapster_match_real_download(user, file, NULL);
      if (real)
	download_close(real, TRANSFER_ERROR);
      return NULL;
   }
   
   temp = g_malloc(sizeof(NAP_DOWNLOAD));
   memset(temp, 0, sizeof(NAP_DOWNLOAD));
   temp->type = N_DOWNLOAD;
   temp->trunc_file = strdup(trunc_file);
   temp->file = strdup(file);
   temp->user = strdup(user);
   temp->id = strdup(id);
   temp->fpath = NULL;
   temp->connection = connection;
   temp->port = port;
   temp->ip = nap_naptohl(ip);
   temp->firewall = alt;
   
   return temp;
}

NAP_SEARCH *gnapster_parse_search(char *s) {
   char *trunc_file, *file, *user, *id, *ip_str;
   int num, size, kbps, freq, connection;
   unsigned long int ip;
   NAP_SEARCH *temp;

   file = next_arg(s, &s);
   id = next_arg(s, &s);
   size = my_atoi(next_arg(s, &s));
   kbps = my_atoi(next_arg(s, &s));
   freq = my_atoi(next_arg(s, &s));
   num = my_atoi(next_arg(s, &s));
   user = next_arg(s, &s);
//   ip = my_atol(next_arg(s, &s));
   ip_str = next_arg(s, &s);
   connection = my_atoi(next_arg(s, &s));
   if (connection < 0) return NULL;
   
   sscanf(ip_str, "%lu", &ip);
   
   trunc_file = strrchr(file, '\\');
   if (!trunc_file) return NULL;
   trunc_file++;
   
   temp = g_malloc(sizeof(NAP_SEARCH));
   memset(temp, 0, sizeof(NAP_SEARCH));
   temp->type = N_SEARCH;
   temp->trunc_file = strdup(trunc_file);
   temp->file = strdup(file);
   temp->user = strdup(user);
   temp->id = strdup(id);
   temp->num = num;
   temp->size = size;
   temp->kbps = kbps;
   temp->freq = freq;
   temp->connection = connection;
   temp->ip = ip;
   
   return temp;
}

NAP_SEARCH *gnapster_parse_resume(char *s) {
   char *trunc_file, *file, *user, *id, *ip_str, *port;
   int size, connection;
   unsigned long int ip;
   NAP_SEARCH *temp;

   user = next_arg(s, &s);
   ip_str = next_arg(s, &s);
   port = next_arg(s, &s);
   file = next_arg(s, &s);
   id = next_arg(s, &s);
   size = my_atoi(next_arg(s, &s));
   connection = my_atoi(next_arg(s, &s));
   if (connection < 0) return NULL;
   
   sscanf(ip_str, "%lu", &ip);
   
   trunc_file = strrchr(file, '\\');
   if (!trunc_file) return NULL;
   trunc_file++;
   
   temp = g_malloc(sizeof(NAP_SEARCH));
   memset(temp, 0, sizeof(NAP_SEARCH));
   temp->type = N_SEARCH;
   temp->trunc_file = strdup(trunc_file);
   temp->file = strdup(file);
   temp->user = strdup(user);
   temp->id = strdup(id);
   temp->size = size;
   temp->connection = connection;
   temp->ip = ip;
   
   return temp;
}

void gnapster_true_fill_resume_path(void *data, int type) {
   FILE *f;
   GList *ptr, *felem;
   char *dptr, *hld;
   char *file, *size, *checksum;
   
   if (!data)
     return;   
   
   f = open_resume("r");
   if (!f)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      file = next_arg_first(dptr, &dptr, &hld);
      size = next_arg(dptr, &dptr);
      checksum = next_arg(dptr, &dptr);
      if (!checksum)
	continue;
      
      if (type == N_DOWNLOAD) {
	 NAP_DOWNLOAD *download = (NAP_DOWNLOAD *)data;
	 
	 if (!strcmp(download->id, checksum)) {
	    if (!download->resume_path)
	      download->resume_path = strdup(file);
	    break;
	 }
      } else {
	 NAP_SEARCH *resume = (NAP_SEARCH *)data;
	 
	 if (!strcmp(resume->id, checksum)) {
	    if (!resume->resume_path)
	      resume->resume_path = strdup(file);
	    break;
	 }
      }
      
      g_free(hld);
   }
   
   g_list_free(felem);
}

void gnapster_fill_dl_resume_path(NAP_DOWNLOAD *download) {
   gnapster_true_fill_resume_path(download, N_DOWNLOAD);
}

void gnapster_fill_resume_path(NAP_SEARCH *resume) {
   gnapster_true_fill_resume_path(resume, N_SEARCH);
}

NAP_SEARCH *gnapster_parse_browse(char *s) {
   char *user, *file, *id;
   int size, kbps, freq, num;
   char *trunc_file;
   NAP_SEARCH *temp;

   user = next_arg(s, &s);
   file = next_arg(s, &s);
   id = next_arg(s, &s);
   size = my_atoi(next_arg(s, &s));
   kbps = my_atoi(next_arg(s, &s));
   freq = my_atoi(next_arg(s, &s));
   num = my_atoi(next_arg(s, &s));
   if (num < 0) return NULL;
   
   trunc_file = strrchr(file, '\\');
   if (!trunc_file) return NULL;
   trunc_file++;
   
   temp = g_malloc(sizeof(NAP_SEARCH));
   memset(temp, 0, sizeof(NAP_SEARCH));
   temp->type = N_SEARCH;
   temp->trunc_file = strdup(trunc_file);
   temp->file = strdup(file);
   temp->user = strdup(user);
   temp->id = strdup(id);
   temp->num = num;
   temp->size = size;
   temp->kbps = kbps;
   temp->freq = freq;
   
   return temp;
}

int gnapster_update_appbar() {
   float val;
   gnapster_main.progress = gnome_appbar_get_progress(GNOME_APPBAR(gnapster_main.appbar));
   val = gtk_progress_get_value(GTK_PROGRESS(gnapster_main.progress));
   val += 0.5;
   if (val > 100) val = 0;
   gtk_progress_set_value(GTK_PROGRESS(gnapster_main.progress), val);
   
   return 1;
}

void gnapster_update_status(char *s, int i) {
   if (s) {
      gnome_appbar_pop(GNOME_APPBAR(gnapster_main.appbar));
      gnome_appbar_push(GNOME_APPBAR(gnapster_main.appbar), s);
   }
   if (i == 0) {
      gtk_timeout_remove(gnapster_main.appbar_timeout);
      gtk_progress_set_activity_mode(GTK_PROGRESS(gnapster_main.progress), 0);
      gtk_progress_set_value(GTK_PROGRESS(gnapster_main.progress), 0);
   } else if (i == 1) {
      if (gnapster_main.appbar_timeout) gtk_timeout_remove(gnapster_main.appbar_timeout);
      gnapster_main.appbar_timeout = gtk_timeout_add(50, gnapster_update_appbar,
						     NULL);
      gtk_progress_set_activity_mode(GTK_PROGRESS(gnapster_main.progress), 1);
      gtk_progress_set_value(GTK_PROGRESS(gnapster_main.progress), 0);
   }
}

void gnapster_highlight(GtkWidget *notebook, GtkWidget *vbox) {
   GtkWidget *label, *notebook_page;
   int curr_page;
   
   label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook),
				      vbox);
   curr_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
   notebook_page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), 
					     curr_page);
   if (notebook_page != GTK_WIDGET(vbox))
     gtk_widget_set_style(label, gnapster_main.highlight);   
}

char *simplify_size(unsigned long int size) {
   float new_size = size;
   int div_count = 0;
   static char buf[25] = { 0 };
   char lbl;
   
   while(new_size >= 1024) {
      new_size /= 1024;
      div_count++;
   }
   
   switch(div_count) {
    case 0:
      lbl = 'b';
      break;
    case 1:
      lbl = 'k';
      break;
    case 2:
      lbl = 'M';
      break;
    case 3:
      lbl = 'G';
      break;
    default:
      lbl = '?';
      break;
   }
   
   g_snprintf(buf, sizeof(buf), "%.02f%c", new_size, lbl);
   
   return buf;
}

int user_lookup(char *user) {
   GList *ptr;
   int i = 0;
   
   ptr = console_tab.channel_list;
   while(ptr) {
      if (!g_strcasecmp(((CHANNEL *)ptr->data)->channel_name, user) &&
	  ((CHANNEL *)ptr->data)->user)
	return (i + 2);
      i++;
      ptr = ptr->next;
   }
   
   return -1;
}

char *text_insert_next_val(char *s, char **next, int *col) {
   char *ptr, *str;
   
   if (!s || !*s) return NULL;
   
   ptr = s;
   str = ptr;
   while(*str) {
      if (*str >= 1 && *str <= 3) {
	 *col = *str;
	 break;
      }
      str++;
   }
   
   if (*str)
     *str++ = 0;
   
   if (next)
     *next = str;
   
   return ptr;
}

char *gnapster_real_text_insert(GtkWidget *text, char *fmt) {
   char *ret = NULL, *fmtPtr, *ptr = NULL;
   int col = -1, next_col = 1;
   GdkColor colors[] = {
      { 0, 0x0000, 0x0000, 0x0000 },
      { 0, 0x0000, 0x0000, 0x7FFF },
      { 0, 0x7FFF, 0x0000, 0x0000 },
      { 0, 0x0000, 0x0000, 0x0000 },
      { 0, 0xaaaa, 0x0000, 0x0000 },
   };
   GdkColor *true_col;
   
   if (!fmt) 
     return ret;
   
   fmtPtr = fmt;
   while(ptr || col == -1) {
      if (col == -1) {
	 col = next_col;
	 ptr = text_insert_next_val(fmtPtr, &fmtPtr, &next_col);
	 ret = ptr;
      } else {
	 col = next_col;
	 ptr = text_insert_next_val(fmtPtr, &fmtPtr, &next_col);
      }
      if (!ptr) break;
      true_col = &colors[col - 1];
      if (user_info.conf[THEME_COLORS]) {
	 if (col == 1) 
	   true_col = &gnapster_main.normal->fg[GTK_STATE_NORMAL];
	 else if (col == 2)
	   true_col = &gnapster_main.highlight->fg[GTK_STATE_NORMAL];
      }
      gtk_text_insert(GTK_TEXT(text), NULL, true_col, NULL, ptr, -1);
   }
   
   return ret;
}

void gnapster_text_insert(int val, char *prefix, char *fmt, ...) {
   GtkWidget *text, *vbox, *parent;
   char *fmtPtr, *format_str;
   va_list args;
   
   if (val == CURR)
     val = gnapster_get_pagenum();
   
   if (val < 0) return;
   
   va_start(args, fmt);
   format_str = g_strdup_vprintf(fmt, args);
   va_end(args);
   if (!format_str)
     return;

   parent = console_tab.notebook;
   switch(val) {
    case CONSOLE:
      text = console_tab.text;
      vbox = console_tab.vbox;
      break;
    case MOTD:
      parent = gnapster_main.notebook;
      text = motd_tab.text;
      vbox = motd_tab.vbox;
      break;
    default:
	{
	   CHANNEL *chan;
	   
	   chan = g_list_nth_data(console_tab.channel_list, (val - 2));
	   if (!chan) 
	     return;

	   text = chan->text;
	   vbox = chan->vbox;
	}
      break;
   }

   if (strcmp(prefix, ""))
     gnapster_text_insert(val, MESSAGE, prefix);
   
   /* fmtPtr will be the first return value from text_insert_next_val, and
    * thus represent the proper memory address */
   fmtPtr = gnapster_real_text_insert(text, format_str);   
   g_free(fmtPtr);
   
   if (parent == console_tab.notebook)
     gnapster_highlight(gnapster_main.notebook, console_tab.main_vbox);
   gnapster_highlight(parent, vbox);
}

void property_free(void *data) {
   WidgetProperty *prop;

   if (!data) return;
   
   prop = (WidgetProperty *)data;
   
   g_free(prop->conf);
   g_list_free(prop->sensitive);
   g_free(prop);
}

void download_free(void *data) {
   NAP_DOWNLOAD *download;
   
   if (!data) return;
   
   download = (NAP_DOWNLOAD *)data;
   
   g_free(download->trunc_file);
   g_free(download->file);
   g_free(download->fpath);
   g_free(download->resume_path);
   g_free(download->user);
   g_free(download);
}

void search_free(void *data) {
   NAP_SEARCH *search;
   
   if (!data) return;
   
   search = (NAP_SEARCH *)data;
   
   g_free(search->trunc_file);
   g_free(search->file);
   g_free(search->user);
   g_free(search->id);
   g_free(search);
}

void server_free(void *data) {
   NAP_SERVER *server;
   
   if (!data) return;
   
   server = (NAP_SERVER *)data;
   
   g_free(server->ip);
   g_free(server->desc);
   g_free(server);
}

void user_free(void *data) {
   NAP_USER *user;
   
   if (!data) return;
   
   user = (NAP_USER *)data;
   
   g_free(user->user);
   g_free(user);
}

void channel_free(void *data) {
   NAP_CHANNEL *channel;
   
   if (!data) return;
   
   channel = (NAP_CHANNEL *)data;
   
   g_free(channel->chan);
   g_free(channel->topic);
   g_free(channel);
}

void resume_free(void *data) {
   NAP_RESUME *resume;
   
   if (!data) return;
   
   resume = (NAP_RESUME *)data;
   
   g_free(resume->file);
   g_free(resume->checksum);
   g_free(resume);
}

void gnapster_destroy(gpointer data) {
   TYPE *n = (TYPE *)data;
   int i = 0;
   
   if (!data) return;
   
   while(gnapster_free[i].func) {
      if (n->type == gnapster_free[i].type) {
	 gnapster_free[i].func (data);
	 break;
      }
      i++;
   }
   if (!gnapster_free[i].func)
     g_free(data);
}

int gnapster_get_pagenum() {       
   int page_num, retval;
   
   page_num = 
     gtk_notebook_get_current_page(GTK_NOTEBOOK(console_tab.notebook));
   
   if (page_num < 0) return -1;
   retval = (page_num == 0) ? 0 : (page_num + 1);
   
   return retval;
}

int gnapster_clist_append(GtkWidget *clist, GtkWidget *vbox, void *data, ...) {
   va_list args;
   GList *ap = NULL;
   char *s;
   int ll, row = -1;

   va_start(args, data);
   while(1) {
      s = va_arg(args, char *);
      if (!s)
	break;
      ap = g_list_append(ap, s);
   }
   va_end(args);
   
   ll = g_list_length(ap);
   
     {
	char *add_id[ll];
	GList *ptr;
	int x = 0;
	
	ptr = ap;
	while(ptr) {
	   add_id[x++] = (ptr->data) ? ptr->data : "???";
	   ptr = ptr->next;
	}

	row = gtk_clist_append(GTK_CLIST(clist), add_id);
   
	gtk_clist_set_row_data_full(GTK_CLIST(clist), row, data,
				    (GtkDestroyNotify)gnapster_destroy);
     }
   

   if (vbox)
     gnapster_highlight(gnapster_main.notebook, vbox);
   
   return row;
}

void *gnapster_get_selection_data(GtkWidget *clist, int *row) {
   void *temp;
   GList *selection;
   
   selection = GTK_CLIST(clist)->selection;
   if (!selection) return NULL;

   temp = gtk_clist_get_row_data(GTK_CLIST(clist), (int)selection->data);
   if (row)
     *row = (int)selection->data;
   
   return temp;
}

int same_color(GdkColor col1, GdkColor col2) {
   if (col1.red == col2.red &&
       col1.green == col2.green &&
       col1.blue == col2.blue)
     return 1;
   
   return 0;
}

void gnapster_get_styles(GtkWidget *widget) {
   gnapster_main.normal = gtk_widget_get_style(widget);
   gnapster_main.highlight = gtk_style_new();
   gnapster_main.highlight->font = gnapster_main.normal->font;
   if (same_color(gnapster_main.normal->fg[GTK_STATE_NORMAL],
		  gnapster_main.normal->fg[GTK_STATE_SELECTED])) {
      gnapster_main.highlight->fg[GTK_STATE_NORMAL].pixel = 
	gnapster_main.normal->fg[GTK_STATE_SELECTED].pixel;
      gnapster_main.highlight->fg[GTK_STATE_NORMAL].red = 65535;
      gnapster_main.highlight->fg[GTK_STATE_NORMAL].green = 0;
      gnapster_main.highlight->fg[GTK_STATE_NORMAL].blue = 0;
   } else
     gnapster_main.highlight->fg[GTK_STATE_NORMAL] = gnapster_main.normal->fg[GTK_STATE_SELECTED];
}

void gnapster_config_set_int(int defaults) {
   char temp_str[1024];
   int i = 0;

   while(conf_table[i].conf_str) {
      g_snprintf(temp_str, sizeof(temp_str), "/gnapster/Options/%s",
		 conf_table[i].conf_str);
      gnome_config_set_int(temp_str, 
			   (defaults == 1) ? conf_table[i].conf_val : user_info.conf[i]);
      i++;
   }
   gnome_config_sync();
}

void gnapster_config_get_int() {
   char temp_str[1024];
   int i = 0;
   
   if (!gnome_config_has_section("/gnapster/Options")) return;
   
   while(conf_table[i].conf_str) {
      g_snprintf(temp_str, sizeof(temp_str), "/gnapster/Options/%s=%i",
		 conf_table[i].conf_str, conf_table[i].conf_val);
      user_info.conf[i] = gnome_config_get_int(temp_str);
      i++;
   }
}

int verify_shared() {
   FILE *f;
   char *ptr, *str, buf[256];
   int args = 0;
   
   f = open_shared(0);
   /* Return 1 to just ignore it */
   if (!f) return 1;
   /* Read the first line only */
   fgets(buf, sizeof(buf) - 1, f);
   fclose(f);
   str = buf;
   do {
      ptr = next_arg(str, &str);
      args++;
   } while(ptr);
   if (args <= 7) return 0;
   
   return 1;
}

void load_signore() {
   char *signore;
   char *ptr = NULL, *freeptr = NULL;
   int i = 0;
   
   signore = gnome_config_get_string("/gnapster/Options/signore");
   while((ptr = next_arg_full(signore, &signore, '\4'))) {
      if (i == 0)
	freeptr = ptr;
      signore_list = g_list_append(signore_list, strdup(ptr));
      i++;
   }
   g_free(freeptr);
}

int idle_startup_cb(gpointer data) {
   if (!strcmp(data, "connect"))
     connect_cb();
   else if (!strcmp(data, "props"))
     preferences_cb();
   return 0;
}

void create_priv_dir() {
   struct stat st;
   char *path;
   
   path = g_strdup_printf("%s/.gnapster", getenv("HOME"));
   if (stat(path, &st) < 0)
     if (mkdir(path, 0755) < 0)
       fprintf(stderr, "Could not create ~/.gnapster directory...functionality may be limited\n");
   
   if (!open_autojoin("r")) {
      FILE *f;
      f = open_autojoin("w");
      if (!f)
	return;

      fprintf(stderr, "autojoin not found, creating defaults...\n");
      
      fprintf(f, "Gnapster\n");
      
      fclose(f);
   }
}

void load_pixmaps() {
   GtkStyle *style;
   
   if (speedgreen != NULL)
     return;
   
   style = gtk_widget_get_style(gnapster_main.window);
   speedgreen = gdk_pixmap_create_from_xpm_d(gnapster_main.window->window,
					     &speedgreenb, &style->bg[GTK_STATE_NORMAL],
					     speedgreen_xpm);
   speedyellow = gdk_pixmap_create_from_xpm_d(gnapster_main.window->window,
					     &speedyellowb, &style->bg[GTK_STATE_NORMAL],
					     speedyellow_xpm);
   speedred = gdk_pixmap_create_from_xpm_d(gnapster_main.window->window,
					     &speedredb, &style->bg[GTK_STATE_NORMAL],
					     speedred_xpm);
   speedgray = gdk_pixmap_create_from_xpm_d(gnapster_main.window->window,
					     &speedgrayb, &style->bg[GTK_STATE_NORMAL],
					     speedgray_xpm);   
}

int main(int argc, char **argv) {
   bindtextdomain(PACKAGE, GNOMELOCALEDIR);
   textdomain(PACKAGE);
   
   gnome_init(PACKAGE, VERSION, argc, argv);

   create_priv_dir();
   gnapster_config_get_int();

   create_gnapster_main();
   
   create_search_tab();
   create_browse_tab();
   create_download_tab();
   create_upload_tab();
   create_console_tab();
   create_motd_tab();
   
   gtk_widget_show_all(gnapster_main.window);
   gnapster_get_styles(gnapster_main.window);
   
   load_pixmaps();

   conn_info.connected = conn_info.connecting =
     conn_info.searching = building = 0;

   memset(&user_info, 0, sizeof(USER_INFO));
   user_info.conf[TERMINATION] = -1;
   
   browse_dlg.window = NULL;
   list_chan_dlg.window = NULL;
   resume_dlg.window = NULL;
   
   load_signore();
   
   user_info.user = gnome_config_get_string("/gnapster/User/user");
   if (!user_info.user) {
      gnapster_update_status(_("Please fill out the preferences form before trying to connect."), -1);
      gtk_idle_add(idle_startup_cb, "props");
   } else {
      gnapster_update_status(_("Auto-connecting..."), -1);
      user_info.pass = gnome_config_get_string("/gnapster/User/pass");
      user_info.download_dir = gnome_config_get_string("/gnapster/User/dl_dir");
      user_info.upload_dir = gnome_config_get_string("/gnapster/User/ul_dir");
      user_info.connection = gnome_config_get_int("/gnapster/User/connection");
      user_info.server = gnome_config_get_string("/gnapster/Options/server");
      if (!gnome_config_has_section("/gnapster/Options")) {
	 user_info.conf[THEME_COLORS] = 0;
	 gnapster_text_insert(CONSOLE, SYSTEM,
			      "Options category not set, assuming defaults.  Please check the preferences dialog for the new preference changes\n");
	 gnapster_config_set_int(1);
      }
      gnapster_config_get_int();
      gtk_idle_add(idle_startup_cb, "connect");
   }

   if (!user_info.conf[FIREWALL]) {
      conn_info.nap_port = gnome_config_get_int("/gnapster/Options/port");
      if (conn_info.nap_port)
	napster_bind(conn_info.nap_port);
      else {
	 gnome_config_set_int("/gnapster/Options/port", NAP_PORT);
	 gnome_config_sync();
	 conn_info.nap_port = NAP_PORT;
	 napster_bind(NAP_PORT);
      }
   } else {
      /* Don't bind */
      conn_info.nap_port = 0;
   }
   
   signal(SIGPIPE, SIG_IGN);

   if (!verify_shared())
     gnome_warning_dialog("Your ~/.gnapster_shared file is outdated, please use the preferences menu to rebuild your list");
   
   gtk_main();
   
   return 1;
}
