/* napster.c - all napster-related protocol stuff */

#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <gnome.h>

#include "napster.h"
#include "gnapster.h"
#include "network.h"
#include "ui.h"
#include "upload.h"
#include "callbacks.h"
#include "queue.h"
#include "regexp.h"
#include "chan.h"

#include "config.h"

#include "commands.h"

extern CONN_INFO conn_info;
extern MOTD_TAB motd_tab;
extern SEARCH_TAB search_tab;
extern BROWSE_TAB browse_tab;
extern CONSOLE_TAB console_tab;
extern USER_INFO user_info;
extern LIST_CHAN_DLG list_chan_dlg;

extern DOWNLOAD_TAB download_tab;
extern UPLOAD_TAB upload_tab;

extern GList *ignore_list;
extern GList *signore_list;
extern GdkPixmap *speedgreen, *speedyellow, *speedred, *speedgray;
extern GdkBitmap *speedgreenb, *speedyellowb, *speedredb, *speedgrayb;

int last_dns = 0;

char *conn_table[] = {
   "Unknown", "14.4", "28.8", "33.6", "56k", "64k ISDN", "128k ISDN",
   "Cable", "DSL", "T1", "T3+", NULL
};

int is_opennap_server(char *data) {
   char *ver, *server, *server_ver;
   int ver_minor, ver_major;

   ver = next_arg(data, &data);
   if (!ver) 
     return 0;
   if (strcmp(ver, "VERSION")) 
     return 0;
   server = next_arg(data, &data);
   if (!server) 
     return 0;
   if (strcmp(server, "opennap")) 
     return 0;
   server_ver = next_arg(data, &data);
   if (!server) 
     return 0;
   ver_major = my_atoi(next_arg_full(server_ver, &server_ver, '.'));
   ver_minor = my_atoi(next_arg_full(server_ver, &server_ver, '.'));
   if (ver_minor < 0) 
     return 0;
   if (ver_major > 0 || (ver_major == 0 && ver_minor >= 12))
     return 1;
   
   return 0;
}

int user_ignored(char *user) {
   GList *ptr;
   char *nick;
   
   if (!user) return 0;
   ptr = ignore_list;
   if (!ptr) return 0;
   while(ptr) {
      nick = ptr->data;
      if (nick) 
	if (!strcmp(nick, user))
	  return 1;
      ptr = ptr->next;
   }
   return 0;
}

void napster_show_links(GtkCTreeNode *node, char *print_str) {
   char *buf;
   char print_str_local[80];
   
   if (print_str) 
     strcpy(print_str_local, print_str);
   else
     print_str_local[0] = '\0';
   
   if (!node)
     node = GTK_CTREE_NODE(GTK_CLIST(console_tab.links_map)->row_list);
   while(node) {
      buf = GTK_CELL_PIXTEXT(GTK_CTREE_ROW(node)->row.cell[0])->text;
      gnapster_text_insert(CURR, SYSTEM,
			   "%s", print_str_local);
      if (GTK_CTREE_ROW(node)->parent)
	gnapster_text_insert(CURR, MESSAGE, 
			     "%s", (GTK_CTREE_ROW(node)->sibling) ? "\2|-- \1" : "\2`-- \1");
      gnapster_text_insert(CURR, MESSAGE, "%s\n", buf);
      if (GTK_CTREE_ROW(node)->children) {
	 if (GTK_CTREE_ROW(node)->parent)
	   strcat(print_str_local,
		  GTK_CTREE_ROW(node)->sibling ? "\2|   \1" : 
		                                 "      ");
	 else
	   strcat(print_str_local, "      ");
	 
	 napster_show_links(GTK_CTREE_ROW(node)->children, print_str_local);
	 print_str_local[strlen(print_str_local) - strlen("      ")] = 0;
      }
      node = GTK_CTREE_ROW(node)->sibling;
   }
}

GtkCTreeNode *napster_find_link(GtkCTreeNode *node, char *s) {
   GtkCTreeNode *work;
   char *buf;
   
   if (!s)
     return NULL;
   
   if (!node)
     node = GTK_CTREE_NODE(GTK_CLIST(console_tab.links_map)->row_list);
   while(node) {
      buf = GTK_CELL_PIXTEXT(GTK_CTREE_ROW(node)->row.cell[0])->text;
      if (!strcmp(buf, s))
	return node;
      if (GTK_CTREE_ROW(node)->children &&
	  (work = napster_find_link(GTK_CTREE_ROW(node)->children, s)))
	return work;
      node = GTK_CTREE_ROW(node)->sibling;
   }
   
   return NULL;
}

void napster_handle_data(int c, char *data) {
   char temp_str[1024];
   
   switch(c) {
    case NAPSTER_MESSAGE:
      g_snprintf(temp_str, sizeof(temp_str), "%s\n", data);
      if (!gtk_text_get_length(GTK_TEXT(motd_tab.text))) {
	 if (is_opennap_server(data)) {
	    //gtk_widget_set_sensitive(search_tab.media, 1);
	    break;
	 }
      }
      gnapster_text_insert(MOTD, MESSAGE, temp_str);
      break;
    case NAPSTER_ADMIN_MESSAGE:
	{
	   char *user, *message;
	   
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   if (!message) break;
	   
	   if (!strcmp(user, user_info.user))
	     break;

	   gnapster_text_insert(CONSOLE, MESSAGE,
				"<\2%s\1/\2wallop\1> \1%s\n", 
				user, message);
	}
      break;
    case NAPSTER_GLOBAL_MESSAGE:
	{
	   char *user, *message;
	   
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   if (!message) break;
	   
	   fprintf(stderr, "[%i] Server notify from %s: %s\n", gnapster_get_pagenum(),
		  user, message);
	   
	   gnapster_text_insert(CONSOLE, MESSAGE,
				"<\2%s\1/\2global\1> %s\n",
				user, message);
	}
      break;
    case 401:
      break;
    case 3:
      g_snprintf(temp_str, sizeof(temp_str), _("Connected (%s:%i): "),
		 conn_info.ip, conn_info.port);
      g_free(conn_info.status);
      conn_info.status = strdup(temp_str);
      if (!conn_info.searching)
	gnapster_update_status(conn_info.status, 0);
      if (!shared_list_handle(user_info.upload_dir, SUBMIT_LIST))
	autojoin_handle();
      break;
    case NAPSTER_STATUS_MESSAGE:
	{
	   int songs, users, gigs;
      
	   users = my_atoi(next_arg(data, &data));
	   songs = my_atoi(next_arg(data, &data));
	   gigs = my_atoi(next_arg(data, &data));
	   if (gigs < 0) break;
	   
	   g_snprintf(temp_str, sizeof(temp_str), 
		      _("Connected (%s:%i): %i songs, %i users, %i %s"),
		      conn_info.ip, conn_info.port, songs, users, gigs,
		      (gigs == 1) ? "gig" : "gigs");
	   if (!conn_info.searching)
	     gnapster_update_status(temp_str, 0);
	   g_free(conn_info.status);
	   conn_info.status = strdup(temp_str);
	}
      break;
    case NAPSTER_USER_MESSAGE:
	{
	   char *user, *message;
	   int index;
	   
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   if (!message) break;
	   
	   if (user_ignored(user)) break;
	   
	   if (user_info.conf[AUTOQUERY])
	     if (user_lookup(user) == -1)
	       do_query(NULL, user);

	   index = user_lookup(user);
	   if (index > 1) {
	      gnapster_text_insert(index, MESSAGE,
				   "<\2%s\1> %s\n", 
				   user, message);
	   } else {
	      gnapster_text_insert(CURR, MESSAGE,
				   "<\2%s\1/\2privmsg\1> %s\n",
				   user, message);
	   }
	}
      break;
    case NAPSTER_CHAN_JOIN:
	{
	   GtkCList *clist;
	   
	   append_console_tab(data, 0);
	   clist = chan_lookup_clist(data);
	   if (!clist) break;
	   
	   gtk_notebook_set_page(GTK_NOTEBOOK(console_tab.notebook),
				 chan_lookup(data) - 1);
	   gtk_clist_freeze(clist);
	}
      break;
    case NAPSTER_CHAN_LIST_ENTRY:
	{
	   char *chan, *users, *topic;
	   NAP_CHANNEL *nchan;
	   
	   chan = next_arg(data, &data);
	   users = next_arg(data, &data);
	   topic = last_arg(data, &data);
	   
	   if (!topic || !list_chan_dlg.window) break;
	   
	   nchan = g_malloc0(sizeof(NAP_CHANNEL));
	   nchan->type = N_CHANNEL;
	   nchan->chan = strdup(chan);
	   nchan->users = my_atoi(users);
	   nchan->topic = strdup(topic);
	   gnapster_clist_append(list_chan_dlg.clist, NULL, nchan,
				 nchan->chan, users, nchan->topic, NULL);
	   gtk_clist_sort(GTK_CLIST(list_chan_dlg.clist));
	}
      break;
    case NAPSTER_CHAN_LIST:
      if (!list_chan_dlg.clist) 
	break;
      gtk_clist_thaw(GTK_CLIST(list_chan_dlg.clist));
      gtk_label_set_text(GTK_LABEL(list_chan_dlg.label), "");
      break;
    case NAPSTER_CHAN_JOINUSER:
    case NAPSTER_CHAN_PARTUSER:
	{
	   char *chan, *user, *sharing;
	   int link_type;
	   GtkCList *clist;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   sharing = next_arg(data, &data);
	   link_type = my_atoi(next_arg(data, &data));
	   if (link_type < 0) break;
	   
	   gnapster_text_insert(chan_lookup(chan), SYSTEM,
				"\2%s\1 %s \2%s\1 (\2%s\1, sharing \2%s\1)\n",
				user, (c == NAPSTER_CHAN_JOINUSER) ? "joined" : "parted",
				chan, napster_conn_as_str(link_type), sharing);
	   clist = chan_lookup_clist(chan);
	   if (!clist) break;
	   
	   if (c == NAPSTER_CHAN_JOINUSER) {
	      NAP_USER *nusr;
	      
	      nusr = g_malloc0(sizeof(NAP_USER));
	      nusr->type = N_USER;
	      nusr->user = strdup(user);
	      nusr->shares = my_atoi(sharing);
	      nusr->conn = link_type;
	      
	      gnapster_clist_append(GTK_WIDGET(clist), NULL, nusr, nusr->user, 
				    sharing, napster_conn_as_str(link_type), 
				    NULL);
	      gtk_clist_sort(clist);
	   } else if (c == NAPSTER_CHAN_PARTUSER) {
	      int i, x = g_list_length(clist->row_list);
	      char *buf;
	      for(i=0; i<x; i++) {
		 gtk_clist_get_text(clist, i, 0, &buf);
		 if (!strcmp(buf, user)) {
		    gtk_clist_remove(clist, i);
		    gtk_clist_sort(clist);
		    break;
		 }
	      }
	   }
	}
      break;
    case NAPSTER_CHAN_USERLIST:
	{
	   char *chan, *user, *share;
	   int conn;
	   NAP_USER *nusr;
	   GtkCList *clist;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   share = next_arg(data, &data);
	   conn = my_atoi(next_arg(data, &data));
	   
	   clist = chan_lookup_clist(chan);
	   
	   if (clist < 0 || !clist) break;

	   nusr = g_malloc0(sizeof(NAP_USER));
	   nusr->type = N_USER;
	   nusr->user = strdup(user);
	   nusr->shares = my_atoi(share);
	   nusr->conn = conn;

	   gnapster_clist_append(GTK_WIDGET(clist), NULL, nusr, nusr->user, 
				    share, napster_conn_as_str(conn),
				    NULL);
	}
      break;
    case NAPSTER_CHAN_USERLIST_END:
	{
	   char *chan;
	   GtkCList *clist;
	   
	   chan = next_arg(data, &data);
	   if (!chan) break;
	   
	   clist = chan_lookup_clist(chan);
	   if (!clist) break;
//	   gtk_clist_sort(clist);
	   gtk_clist_thaw(clist);
	}
      break;
    case NAPSTER_CHAN_TOPIC:
	{
	   char *chan, *topic;

	   chan = next_arg(data, &data);
	   topic = last_arg(data, &data);
	   if (!topic) break;
	   
	   gnapster_text_insert(chan_lookup(chan), SYSTEM, "Topic [\2%s\1]: \2%s\1\n",
				chan, topic);
	}
      break;
    case NAPSTER_CHAN_MESSAGE_IN:
	{
	   char *chan, *user, *message;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   if (!message) break;
	   
	   if (user_ignored(user)) break;
	   
	   if (strcmp(user, user_info.user)) {
	      char *al1, *al2, c1, c2;

	      al1 = g_strdup(all_lower(message));
	      al2 = g_strdup(all_lower(user_info.user));
	      if (!al1 || !al2)
		break;
	      
	      c1 = (strstr(al1, al2)) ? '(' : '<';
	      c2 = (c1 == '(') ? ')' : '>';
	      
	      g_free(al1);
	      g_free(al2);

	      gnapster_text_insert(chan_lookup(chan), MESSAGE,
				   "%c\2%s\1%c %s\n", c1, user, c2, message);
	   }
	}
      break;
    case NAPSTER_LINKS:
	{
	   char *host1, *host2, *text[2], *hops;
	   int port1, port2;
	   GtkCTreeNode *parent;
	   
	   if (!data || !data[0]) {
	      if (console_tab.links_map) {
		 napster_show_links(NULL, NULL);
		 gtk_widget_destroy(console_tab.links_map);
	      } else
		gnapster_text_insert(CURR, SYSTEM,
				     "No links found\n");
	      console_tab.links_map = NULL;
	      break;
	   }
	   
	   host1 = next_arg(data, &data);
	   port1 = my_atoi(next_arg(data, &data));
	   host2 = next_arg(data, &data);
	   port2 = my_atoi(next_arg(data, &data));
	   hops = next_arg(data, &data);
	   
	   if (!hops)
	     break;

	   if (!console_tab.links_map)
	     console_tab.links_map = gtk_ctree_new(2, 0);
	   
	   parent = napster_find_link(NULL, host1);
	   if (!parent) {
	      text[0] = host1;
	      text[1] = "-1";
	      text[2] = NULL;
	      gtk_ctree_insert_node(GTK_CTREE(console_tab.links_map),
				    NULL, NULL, text, 0, NULL, NULL, NULL,
				    NULL, 0, 0);
	      
	      /* Look again now */
	      parent = napster_find_link(NULL, host1);
	   }
	   
	   text[0] = host2;
	   text[1] = hops;
	   text[2] = NULL;
	   
	   gtk_ctree_insert_node(GTK_CTREE(console_tab.links_map),
				 parent, NULL, text, 0, NULL, NULL, NULL, 
				 NULL, 0, 0);
	   
/*	   gnapster_text_insert(gnapster_get_pagenum(), "%s\2:\1%i \2->\1 %s\2:\1%i (%s)\n",
				host1, port1, host2, port2, hops);*/
	}
      break;
    case NAPSTER_WHOIS_REPLY:
	{
	   int i;
	   char *values[10], temp_val[1024];
	   char *value_table[] = {
              "User", "Account", "Idle Time", "Channels","Status", "Sharing", 
	      "Downloads", "Uploads", "Connection", "Client", "Total Uploads",
	      "Total Downloads", "IP", "Connect port", "Data port", "Email", 
	      NULL
	   };

	   gnapster_text_insert(CURR, MESSAGE,
				"\2**********************************************************\n");
	   for(i=0; i<=sizeof(value_table); i++) {
	      values[i] = next_arg(data, &data);
	      if (!values[i]) break;
	      if (!strcmp(value_table[i], "Idle Time")) {
		 int idle_time = atoi(values[i]);
		 g_snprintf(temp_val, sizeof(temp_val), "%i\1 m \2%i\1 s",
			    idle_time / 60, idle_time % 60);
	      } else if (!strcmp(value_table[i], "Connection")) {
		 int conn = atoi(values[i]);
		 g_snprintf(temp_val, sizeof(temp_val), 
			    napster_conn_as_str(conn));
	      } else {
		 g_snprintf(temp_val, sizeof(temp_val), values[i]);
	      }
	      gnapster_text_insert(CURR, MESSAGE, "%s\1 %s: \2%s\1%c",
				   (i % 2 == 0) ? "\2*" : "\1,", value_table[i],
				   (strcmp(temp_val, " ") && strcmp(temp_val, "")) ? temp_val : "(none)",
				   (i % 2 == 0) ? '\2' : '\n');
	   }
	   gnapster_text_insert(CURR, MESSAGE,
				"\2**********************************************************\n");
	}
      break;
    case NAPSTER_WHOIS_OFFLINE:
	{
	   char *user, *account;

	   user = next_arg(data, &data);
	   account = next_arg(data, &data);
	   if (!account) break;

	   gnapster_text_insert(CURR, SYSTEM, "\2%s \1is offline.\n",
				user);
	}
      break;
    case NAPSTER_UPLOAD_REQUEST:
	{
	   char *user, *file, *trunc_file;
	   
	   user = next_arg(data, &data);
	   file = next_arg(data, &data);
	   if (!file) break;
	   trunc_file = strrchr(file, '\\');
	   if (!trunc_file) break;
	   trunc_file++;

	   if (user_ignored(user))
	     break;

	   if (!upload_accepted(user, file, trunc_file))
	     break;
	   
	   napster_send(conn_info.sock, NAPSTER_ACCEPT_UPLOAD_REQUEST,
			"%s \"%s\"", user, file);
	}
      break;
    case NAPSTER_SERVER_MESSAGE:
	{
	   GList *ptr;
	   char *str;
	   int match = 0;
	   
	   for(ptr=signore_list; ptr; ptr=ptr->next) {
	      str = ptr->data;
	      if (str) {
		 if (wild_match(str, data))
		   match = 1;
	      }
	      if (match)
		break;
	   }
	   
	   if (!strncmp(data, "duplicate file", strlen("duplicate file")))
	     match = 1;
	   
	   if (!match)
	     gnapster_text_insert(CURR, SYSTEM, "[\2server\1]: %s\n", data);
	   
	   match = 0;
	   if (wild_match("User % is not * online*", data))
	     match = 1;
	   
	   if (match) {
	      gnapster_update_status(conn_info.status, 0);
	      gtk_entry_set_editable(GTK_ENTRY(search_tab.gtk_entry), 1);
	      gtk_entry_set_editable(GTK_ENTRY(browse_tab.entry), 1);
	      conn_info.searching = 0;
	   }
	}
      break;
    case NAPSTER_RESUME_RESULT:
	{
	   NAP_SEARCH *resume;
	   
	   resume = gnapster_parse_resume(data);
	   gnapster_fill_resume_path(resume);
	   
	   gnapster_clist_append(search_tab.clist, search_tab.vbox,
				 resume, resume->trunc_file, 
				 napster_conn_as_str(resume->connection),
				 resume->user, "", "", NULL);
	}
      break;
    case NAPSTER_RESUME_END:
      gnapster_update_status(conn_info.status, 0);
      //gtk_entry_set_text(GTK_ENTRY(search_tab.gtk_entry), "");
      gtk_entry_set_editable(GTK_ENTRY(search_tab.gtk_entry), 1);
      gtk_clist_sort(GTK_CLIST(search_tab.clist));
      gtk_clist_thaw(GTK_CLIST(search_tab.clist));
      if (!g_list_length(GTK_CLIST(search_tab.clist)->row_list))
	gnapster_clist_append(search_tab.clist, search_tab.vbox, NULL,
			     "No resume entries found", "", "", "", "", "", NULL);
      conn_info.searching = 0;
      break;
    case NAPSTER_SEARCH_RESULT:
	{
	   NAP_SEARCH *search;
	   char temp_str2[10], temp_str3[5];
	   int seconds, row;
	   GdkPixmap *pixmap = NULL;
	   GdkBitmap *bitmap = NULL;
	   
	   search = gnapster_parse_search(data);
	   if (!search) break;
	   
	   if (search->kbps)
	     seconds = (search->size / ((search->kbps * 1000) / 8));
	   else
	     seconds = 0;
	   
	   g_snprintf(temp_str, sizeof(temp_str), "%.02fM", ((float)search->size / 1024) / 1024);
	   g_snprintf(temp_str2, sizeof(temp_str2), "%i:%02i",
		      seconds / 60, seconds % 60);
	   g_snprintf(temp_str3, sizeof(temp_str3), "%i", search->kbps);
	   row = gnapster_clist_append(search_tab.clist, search_tab.vbox,
				       search, search->trunc_file,
				       napster_conn_as_str(search->connection),
				       search->user, temp_str, temp_str2, temp_str3, NULL);
	   
	   if (search->connection > 10 || search->connection < 1)
	     pixmap = speedgray, bitmap = speedgrayb;
	   else if (search->connection <= 4)
	     pixmap = speedred, bitmap = speedredb;
	   else if (search->connection <= 6)
	     pixmap = speedyellow, bitmap = speedyellowb;
	   else if (search->connection <= 10)
	     pixmap = speedgreen, bitmap = speedgreenb;
	   else 
	     pixmap = speedgray, bitmap = speedgrayb;
	   
	   gtk_clist_set_pixtext(GTK_CLIST(search_tab.clist), row, 0,
				 search->trunc_file, 3, pixmap, bitmap);
	}
      break;
    case NAPSTER_END_SEARCH:
      g_snprintf(temp_str, sizeof(temp_str), "Nothing found for '%s'",
		 gtk_entry_get_text(GTK_ENTRY(search_tab.gtk_entry)));
      gnapster_update_status(conn_info.status, 0);
      gtk_entry_set_text(GTK_ENTRY(search_tab.gtk_entry), "");
      gtk_entry_set_editable(GTK_ENTRY(search_tab.gtk_entry), 1);
      gtk_clist_sort(GTK_CLIST(search_tab.clist));
      gtk_clist_thaw(GTK_CLIST(search_tab.clist));
      if (!g_list_length(GTK_CLIST(search_tab.clist)->row_list))
	gnapster_clist_append(search_tab.clist, search_tab.vbox, NULL,
			      temp_str, "", "", "", "", "", NULL);
      conn_info.searching = 0;
      break;
    case NAPSTER_BROWSE_RESULT:
	{
	   NAP_SEARCH *browse;
	   
	   browse = gnapster_parse_browse(data);
	   if (!browse) break;

	   browse_tab_add(browse);
	}
      break;
    case NAPSTER_END_BROWSE:
      gnapster_update_status(conn_info.status, 0);
      gtk_entry_set_text(GTK_ENTRY(browse_tab.entry), "");
      gtk_entry_set_editable(GTK_ENTRY(browse_tab.entry), 1);
      gtk_ctree_sort_recursive(GTK_CTREE(browse_tab.ctree), NULL);
      gtk_clist_thaw(GTK_CLIST(browse_tab.ctree));
      if (g_list_length(GTK_CLIST(browse_tab.ctree)->row_list) == 0)
	gnapster_clist_append(browse_tab.ctree, browse_tab.vbox, NULL,
			      "No mp3s found.", "", "", "", "", NULL);
      conn_info.searching = 0;
      break;
    case NAPSTER_BROWSE_ERROR:
      gnapster_text_insert(CONSOLE, SYSTEM, "User is not online\n");
      gnapster_update_status(conn_info.status, 0);
      gtk_clist_thaw(GTK_CLIST(browse_tab.ctree));
      conn_info.searching = 0;
      break;
    case NAPSTER_DOWNLOAD_REPLY:
	{
	   NAP_DOWNLOAD *download, *real;
	   int row = -1;
	   
	   download = gnapster_parse_download(data, 0);
	   if (!download) 
	     break;
	   
	   DPRINTF(("Received download reply from %s (for %s)\n",
		    download->user, download->file));

	   real = gnapster_match_real_download(download->user, download->file, &row);
	   if (!real) {
	      download_free(download);
	      break;
	   }
	   if (real->status != 0) {
	      download_free(download);
	      break;
	   }
	   
	   if (download->port == 0) {
	      napster_send(conn_info.sock, NAPSTER_ALT_GET, "%s \"%s\"",
			   download->user, download->file);
	      gtk_clist_set_text(GTK_CLIST(download_tab.clist), row, 3,
				 "Awaiting connection...");
	      real->status++;
	      download_free(download);
	   } else {
	      g_free(real->trunc_file);
	      g_free(real->file);
	      g_free(real->user);
	      real->user = strdup(download->user);
	      real->ip = download->ip;
	      real->port = download->port;
	      real->file = strdup(download->file);
	      real->trunc_file = strdup(download->trunc_file);
	      real->status++;
	      gtk_clist_set_text(GTK_CLIST(download_tab.clist),
				 row, 3, "Connecting...");
	      download_free(download);
	      napster_connect_user(real);
	   }
	}
      break;
    case NAPSTER_ALT_DOWNLOAD_REPLY:
	{
	   NAP_DOWNLOAD *download;
	   
	   download = gnapster_parse_download(data, 1);
	   if (!download) break;
	   
	   DPRINTF(("Received alternate download reply from %s...\n",
		    download->user));
	   
	   /* Ok, we need to connect to them now */
	   if (download->port == 0) {
	      gnapster_text_insert(CONSOLE, SYSTEM, "Unable to send to \2%s\1, user cannot accept connections\n",
				   download->user);
	      download_free(download);
	   } else
	     napster_connect_user(download);
	}
      break;
    case NAPSTER_DOWNLOAD_ERROR:
	{
	   char *user, *file;
	   NAP_DOWNLOAD *real;
	   
	   user = next_arg(data, &data);
	   file = next_arg(data, &data);
	   if (!file) break;

	   gnapster_text_insert(CONSOLE, SYSTEM, "\2%s\1 has disconnected.\n",
				user);
	   real = gnapster_match_real_download(user, file, NULL);
	   download_close(real, TRANSFER_ERROR);
	}
      break;
    case NAPSTER_SENDLIMIT_REACHED:
	{
	   char *user, *file, *size_str, *trunc_file;
	   unsigned long int size;
	   int limit, row = -1;
	   NAP_DOWNLOAD *real;
	   
	   user = next_arg(data, &data);
	   file = next_arg(data, &data);
	   size_str = next_arg(data, &data);
	   limit = my_atoi(next_arg(data, &data));
	   if (limit < 0) break;
	   sscanf(size_str, "%lu", &size);
	   trunc_file = strrchr(file, '\\');
	   if (!trunc_file) break;
	   trunc_file++;
	   
	   real = gnapster_match_real_download(user, file, &row);
	   if (!real)
	     break;
	   gtk_clist_set_text(GTK_CLIST(download_tab.clist), row, 3,
			      "Remotely queued");
	   /* Try again every 3 seconds */
	   if (!real->timeout_tag) {
	      if (real->start_timeout)
		gtk_timeout_remove(real->start_timeout);
	      real->start_timeout = gtk_timeout_add(60000  * 7,
						    network_download_start_timeout, real);
	      real->timeout_tag = gtk_timeout_add(60000 * 3,
						  remote_queue_check, real);
	   }
	   real->status = 1;
//	   download_close(real, TRANSFER_ERROR);
	}
      break;
    case NAPSTER_PING:
	{
	   char *user;
	   
	   user = next_arg(data, &data);
	   if (!user) break;
	   
	   if (user_ignored(user))
	     break;
	   
	   gnapster_text_insert(CONSOLE, SYSTEM, "\2%s\1 has pinged you\n",
				user);
	   napster_send(conn_info.sock, NAPSTER_PONG, user);
	}
      break;
    case NAPSTER_ACTION:
	{
	   char *chan, *user, *message;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   /* it's in quotes */
	   message = next_arg(data, &data);
	   if (!message) break;
	   
	   if (user_ignored(user) ||
	       !strcmp(user, user_info.user))
	     break;
	   
	   gnapster_text_insert(chan_lookup(chan), SYSTEM, "\2%s\1 %s\n", user, message);
	}
      break;
    case NAPSTER_REQUEST_PORTCHANGE:
	{
	   int port;
	   
	   port = my_atoi(next_arg(data, &data));
	   if (port < 0) 
	     break;
	   
	   if (conn_info.nap_port != port) {
	      gnapster_text_insert(CONSOLE, SYSTEM, "\2Warning\1: Server is attempting to change your data port to \2%i\1\n",
				   port);
	      gdk_input_remove(conn_info.bind_sock_input);
	      close(conn_info.bind_sock);
	      if (port)
		napster_bind(port);
	      conn_info.nap_port = port;
	   }
	}
      break;
    case NAPSTER_NEWUSER_SUCCESS:
      break;
    case NAPSTER_NEWUSER_REG:
      gnome_error_dialog(_("The username you have chosen has already been taken on Gnapster.  Please choose a new one."));
      disconnect_cb();
      break;
    case NAPSTER_NEWUSER_INVALID_NICK:
      gnome_error_dialog(_("Username considered invalid by server.  Please try again"));
      disconnect_cb();
      break;
    case NAPSTER_ERROR:
      gnome_error_dialog(data);
      fprintf(stderr, "error -> %s\n", data);
      break;
    default:
      fprintf(stderr, "unknown[%i] -> %s\n", c, data);
      break;
   }
}

/* Not being used, couldn't figure it out */
void napster_dns_cb(guint32 ip_addr, int *sock) {
   struct sockaddr_in server;
   
   server.sin_family = AF_INET;
   server.sin_port = htons(8875);
   server.sin_addr.s_addr = htonl(ip_addr);
   
   fprintf(stderr, "napster_dns_cb -> %s\n", inet_ntoa(server.sin_addr));
   
   fcntl(*sock, F_SETFL, O_NONBLOCK);
   connect(*sock, (struct sockaddr *)&server, sizeof(server));
   
   conn_info.sock_input = gdk_input_add(*sock, GDK_INPUT_WRITE,
					network_conn_cb, NULL);
   
}

void napster_get_best_host() {
   int sock;
   struct sockaddr_in server;
   
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
   if (sock < 0) {
      fprintf(stderr, "socket error\n");
      /* Error handle */
      return;
   }
   
   server.sin_family = AF_INET;

   g_free(user_info.server);
   user_info.server = gnome_config_get_string("/gnapster/Options/server");
   user_info.server_port = gnome_config_get_int("/gnapster/Options/server_port");
   
   server.sin_addr.s_addr = inet_addr(user_info.server);
   server.sin_port = htons(user_info.server_port);
   if (server.sin_addr.s_addr == -1) {
      gnome_error_dialog("Server is likely in hostname format, as of 1.3.7 IP addresses are required.\nPlease see the browse servers dialog to assist you in connecting.");
      disconnect_cb();
      return;
   }
   
   fcntl(sock, F_SETFL, O_NONBLOCK);
   connect(sock, (struct sockaddr *)&server, sizeof(server));
   
   conn_info.sock_input = gdk_input_add(sock, GDK_INPUT_WRITE,
					network_conn_cb, NULL);
}

void napster_connect(char *ip, int port) {
   int sock;
   struct sockaddr_in server;
   
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
   if (sock < 0) {
      /* error check */
      fprintf(stderr, "bogus\n");
      return;
   }
   
   server.sin_family = AF_INET;
   server.sin_port = htons(port);
   server.sin_addr.s_addr = inet_addr(ip);
/*   conn_info.saddr.sin_family = AF_INET;
   conn_info.saddr.sin_port = htons(port);
   conn_info.saddr.sin_addr.s_addr = inet_addr(ip);*/
   
   fcntl(sock, F_SETFL, O_NONBLOCK);
   connect(sock, (struct sockaddr *)&server, sizeof(server));
   
   conn_info.sock_input = gdk_input_add(sock, GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
					network_conn_real_cb, (struct sockaddr_in *)&server);
}

void napster_connect_user(NAP_DOWNLOAD *download) {   
   int sock;
   
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
   if (sock < 0) {
      /* error check */
      fprintf(stderr, "bogus\n");
      return;
   }
   
   DPRINTF(("download = %lu:%i\n", download->ip, download->port));
   
   download->saddr.sin_family = AF_INET;
   download->saddr.sin_port = htons(download->port);
   download->saddr.sin_addr.s_addr = download->ip;
   
   fcntl(sock, F_SETFL, O_NONBLOCK);
   connect(sock, (struct sockaddr *)&download->saddr, sizeof(download->saddr));
   
   download->sock = sock;
   download->header_val = 0;
   download->size = 0;
   download->sock_input = gdk_input_add(sock, GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
					network_conn_user_cb, download);
}

void napster_bind(int port) {
   int sock, len = 1;
   struct sockaddr_in server;
   
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
   if (sock < 0) {
      fprintf(stderr, "socket error!\n");
      return;
   }
   
   server.sin_family = AF_INET;
   server.sin_addr.s_addr = htonl(INADDR_ANY);
   server.sin_port = htons(port);
   
   setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &len, sizeof(len));
   
   if (bind(sock, (struct sockaddr *)&server, sizeof(server)) < 0) {
      close(sock);
      perror("bind");
      gnome_error_dialog(_("An error occured while attempting to bind port 6699.  It is possible that another napster client is running."));
      return;
   }
   listen(sock, 10 + user_info.conf[MAX_UPLOADS]);
   
   conn_info.bind_sock = sock;
   conn_info.bind_sock_input = gdk_input_add(conn_info.bind_sock,
					     GDK_INPUT_READ,
					     network_handle_connection,
					     NULL);
   
   return;
}

void napster_send_get(int sock, char *user, char *file, int conn) {
   char *temp_str;
   
   send(sock, "GET", 3, 0);
   temp_str = g_strdup_printf("%s \"%s\" %i", user, file, conn);
   send(sock, temp_str, strlen(temp_str), 0);
   g_free(temp_str);
}

void napster_send_send(int sock, char *user, char *file, unsigned long int size) {
   char *temp_str;
   
   send(sock, "SEND", 4, 0);
   temp_str = g_strdup_printf("%s \"%s\" %lu", user, file, size);
   send(sock, temp_str, strlen(temp_str), 0);
   g_free(temp_str);
}

int napster_send(int sock, int cmd, char *fmt, ...) {
   PACKET *p;
   va_list args;
   char *data = NULL;
   int x;
   
   if (cmd < 0 || sock < 0) return -1;
   
   if (fmt) {
      va_start(args, fmt);
      data = g_strdup_vprintf(fmt, args);
      va_end(args);
   }
   
   p = g_malloc0(sizeof(PACKET));

   if (data) 
     p->len = strlen(data);
   else p->len = 0;
   p->cmd = cmd;
   p->len = BSWAP16(p->len);
   p->cmd = BSWAP16(p->cmd);
      
   x = send(sock, p, 4, 0);
   if (data)
     x = send(sock, data, strlen(data), 0);
   
   g_free(data);
   g_free(p);
   
   return x;
}

int napster_conn_as_int(char *s) {
   int x = 0;
   
   while(conn_table[x]) {
      if (!strcmp(conn_table[x], s)) return x;
      x++;
   }
   
   return -1;
}

char *napster_conn_as_str(int i) {
   if (i < 0 || i > 10) return "Unknown";
   return conn_table[i];
}

int napster_user_uploads(char *user) {
   GList *ptr;
   GtkCList *clist;
   GtkCListRow *clist_row;
   NAP_DOWNLOAD *data;
   int uploads = 0;
   
   clist = GTK_CLIST(upload_tab.clist);
   ptr = clist->row_list;
   if (!ptr)
     return -1;
   while(ptr) {
      clist_row = ptr->data;
      data = clist_row->data;
      if (!data)
	break;
      if (!strcmp(data->user, user))
	uploads++;
      ptr = ptr->next;
   }
   
   return uploads;
}
