#ifndef __NAPSTER_H
#define __NAPSTER_H

#define NAPSTER_ERROR 0

#define NAPSTER_LOGIN 2
#define NAPSTER_NEW_USER 7
#define NAPSTER_NEWUSER_SUCCESS 8
#define NAPSTER_NEWUSER_REG 9
#define NAPSTER_NEWUSER_INVALID_NICK 10

#define NAPSTER_SUBMIT 100
#define NAPSTER_REMOVE_SHARE 102

#define NAPSTER_SEARCH 200
#define NAPSTER_SEARCH_RESULT 201
#define NAPSTER_END_SEARCH 202

#define NAPSTER_GET 203
#define NAPSTER_DOWNLOAD_REPLY 204

#define NAPSTER_USER_MESSAGE 205

#define NAPSTER_DOWNLOAD_ERROR 206

#define NAPSTER_BROWSE_ERROR 210
#define NAPSTER_BROWSE 211
#define NAPSTER_BROWSE_RESULT 212
#define NAPSTER_END_BROWSE 213

#define NAPSTER_STATUS_MESSAGE 214

#define NAPSTER_RESUME_SEARCH 215
#define NAPSTER_RESUME_RESULT 216
#define NAPSTER_RESUME_END 217

#define NAPSTER_DOWNLOAD_START 218
#define NAPSTER_DOWNLOAD_COMPLETE 219
#define NAPSTER_UPLOAD_START 220
#define NAPSTER_UPLOAD_COMPLETE 221

#define NAPSTER_JOIN_CHAN 400
#define NAPSTER_PART_CHAN 401
#define NAPSTER_CHAN_MESSAGE 402
#define NAPSTER_CHAN_MESSAGE_IN 403

#define NAPSTER_SERVER_MESSAGE 404

#define NAPSTER_CHAN_JOIN 405
#define NAPSTER_CHAN_JOINUSER 406
#define NAPSTER_CHAN_PARTUSER 407
#define NAPSTER_CHAN_USERLIST 408
#define NAPSTER_CHAN_USERLIST_END 409
#define NAPSTER_CHAN_TOPIC 410

#define NAPSTER_ALT_GET 500
#define NAPSTER_ALT_DOWNLOAD_REPLY 501 

#define NAPSTER_WHOIS_REQUEST 603
#define NAPSTER_WHOIS_REPLY 604
#define NAPSTER_WHOIS_OFFLINE 605

#define NAPSTER_CHANGE_USERLEVEL 606

#define NAPSTER_UPLOAD_REQUEST 607
#define NAPSTER_ACCEPT_UPLOAD_REQUEST 608

#define NAPSTER_KILL_USER 610
#define NAPSTER_BAN_USER 612
#define NAPSTER_UNBAN_USER 614

#define NAPSTER_REQUEST_PORTCHANGE 613

#define NAPSTER_CHAN_LIST 617
#define NAPSTER_CHAN_LIST_ENTRY 618

#define NAPSTER_SENDLIMIT 619
#define NAPSTER_SENDLIMIT_REACHED 620

#define NAPSTER_MESSAGE 621
#define NAPSTER_MUZZLE_USER 622
#define NAPSTER_UNMUZZLE_USER 623

#define NAPSTER_ADMIN_MESSAGE 627
#define NAPSTER_GLOBAL_MESSAGE 628

#define NAPSTER_CHANGE_PORT 703

#define NAPSTER_PING 751
#define NAPSTER_PONG 752

#define NAPSTER_SERVER_VERSION 801

#define NAPSTER_ACTION 824

#define NAPSTER_CHANNELLIMIT 826

/* Non-standard opennap command */
#define NAPSTER_LINKS 10112


#define N_DOWNLOAD 0
#define N_SEARCH 1
#define N_SERVER 2
#define N_FILE 3
#define N_USER 4
#define N_CHANNEL 5
#define N_RESUME 6
#define W_PROP 7

typedef struct _nap_download {
   int type;
   char *trunc_file;
   char *file;
   char *user;
   int connection;
   int port;
   unsigned long int ip;
   
   struct sockaddr_in saddr;
   
   int sock;
   int sock_input;
   int update_tag;
   int accept_timeout;
   int start_timeout;
   FILE *fdesc;
   char *fpath;
   char header[4096];
   int header_val;
   int size, temp_size;
   unsigned long int total;
   int calls;
   float kps;
   int timeout_calls;
   int firewall;
   int flag;
   int upload;
   
   int status;
   int timeout_tag;
   
   /* Data it gets from a nap_search */
   char *id;
   int num;
   int kbps;
   int freq;
   char *resume_path;
} NAP_DOWNLOAD;

typedef struct _nap_search {
   int type;
   char *file;
   char *trunc_file;
   char *id;
   int size;
   int kbps;
   int freq;
   int num;
   char *user;
   unsigned long int ip;
   int connection;
   char *resume_path;
} NAP_SEARCH;

typedef struct _nap_server {
   int type;
   char *ip;
   int port;
   char *desc;
} NAP_SERVER;

typedef struct _nap_file {
   int type;
   time_t mtime;
   char *filename;
} NAP_FILE;

typedef struct _nap_user {
   int type;
   char *user;
   int shares;
   int conn;
} NAP_USER;

typedef struct _nap_channel {
   int type;
   char *chan;
   int users;
   char *topic;
} NAP_CHANNEL;

typedef struct _nap_resume {
   int type;
   char *file;
   char *checksum;
   int localsize;
   int remotesize;
} NAP_RESUME;

int user_ignored(char *);
void napster_dns_cb(guint32, int *);
void napster_get_best_host();
void napster_connect_cddb();
void napster_connect(char *, int);
void napster_connect_user(NAP_DOWNLOAD *);
void napster_bind(int);
void napster_send_get(int, char *, char *, int);
void napster_send_send(int, char *, char *, unsigned long int);
int napster_send(int, int, char *, ...);
int is_opennap_server(char *);
void napster_handle_data(int, char *);
int napster_conn_as_int(char *);
char *napster_conn_as_str(int);
int napster_user_uploads(char *);

#endif /* __NAPSTER_H */
