/* network.c - general socket and network code callbacks  */

#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <ctype.h>

#include <gnome.h>

#include "napster.h"
#include "network.h"
#include "gnapster.h"
#include "ui.h"
#include "upload.h"
#include "callbacks.h"
#include "queue.h"
#include "servers.h"
#include "resume.h"
#include "chan.h"

#include "config.h"

#define CLIST_UPDATE_SPEED 1000

extern CONN_INFO conn_info;
extern USER_INFO user_info;
extern DOWNLOAD_TAB download_tab;
extern UPLOAD_TAB upload_tab;
extern CONSOLE_TAB console_tab;
extern SEARCH_TAB search_tab;
extern MOTD_TAB motd_tab;
extern BROWSE_DLG browse_dlg;

char rd[1024];
int rd_val = 0, rd_header = 0;

/* This data is used to gather a packet w/o blocking the GTK
 * code */
int packet_len, packet_cmd, total_packet_len, read_count;
char packet_data[1024 * 4];

extern GList *total_list;

struct in_addr *network_atoaddr(char *address) {
   struct hostent *host;
   static struct in_addr saddr;
   
   saddr.s_addr = inet_addr(address);
   if (saddr.s_addr != -1) return &saddr;
   host = gethostbyname(address);
   if (host != NULL) return (struct in_addr *)*host->h_addr_list;
   return NULL;
}

void network_read_napigator(gpointer data, int source, GdkInputCondition condition) {
   int n, i;
   char buf[1024];
   
   n = recv(source, buf, sizeof(buf) - 1, 0);
   if (n <= 0) {
//      perror("recv");
      refresh_dlg_cb(NULL, NULL);
//      gdk_input_remove(conn_info.napigator_input);
//      close(source);
      return;
   }
   buf[n] = 0;
   
   for(i=0; i<n; i++) {
      if (buf[i] == '\n') {
	 char *ptr;
	 char *ip, *port, *network, *title;
	 
	 if (rd_val && rd[rd_val - 1] == '\r')
	   rd[rd_val -1] = 0;
	 
	 rd[rd_val] = 0;
	 rd_val = 0;
	 
	 if (!strcmp(rd, "Content-Type: text/html"))
	   rd_header = 1;
	 else if (rd_header) {
	    ptr = rd;
	 
	    ip = next_arg(ptr, &ptr);
	    port = next_arg(ptr, &ptr);
	    network = next_arg(ptr, &ptr);
	    next_arg(ptr, &ptr), next_arg(ptr, &ptr), next_arg(ptr, &ptr);
	    title = next_arg(ptr, &ptr);
	    if (network && !strcmp(network, "Napster")) {
//	       gtk_clist_freeze(GTK_CLIST(browse_dlg.clist));
//	       gtk_clist_clear(GTK_CLIST(browse_dlg.clist));
	       populate_browse_dlg_clist();
//	       gtk_clist_thaw(GTK_CLIST(browse_dlg.clist));

	       refresh_dlg_cb(NULL, NULL);
	       gdk_input_remove(conn_info.napigator_input);
	       close(source);
	       return;
	    } else if (network) {
	       if (!check_server_match(ip, port))
		 append_server_entry(ip, port, title);
	    }
	 }
      } else rd[rd_val++] = buf[i];
   }
}

void network_conn_napigator_cb(gpointer data, int source, GdkInputCondition condition) {
   int err, len;
   
   gdk_input_remove(conn_info.napigator_input);
   
   len = sizeof(err);
   getsockopt(source, SOL_SOCKET, SO_ERROR, &err, &len);
   if (err) {
      gnome_error_dialog("Could not create connection to www.napigator.com");
      return;
   }
   
   fcntl(source, F_SETFL, 0);
   
   send(source, "GET /servers.php?version=107&client=gnapster HTTP/1.0\n\n", 
	strlen("GET /servers.php?version=107&client=gnapster HTTP/1.0\n\n"), 0);
   
   rd_header = 0;
   conn_info.napigator_input = gdk_input_add(source, GDK_INPUT_READ,
					     network_read_napigator, NULL);
}

void network_connect_napigator() {
   int sock;
   struct sockaddr_in server;
   
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
   if (sock < 0) {
      perror("socket");
      return;
   }
   
   //memset(&server, 0, sizeof(server));
   server.sin_family = AF_INET;
   server.sin_port = htons(80);
   server.sin_addr.s_addr = inet_addr("216.65.92.23");
   
   fcntl(sock, F_SETFL, O_NONBLOCK);
   connect(sock, (struct sockaddr *)&server, sizeof(server));
   
   conn_info.napigator_sock = sock;
   conn_info.napigator_input = gdk_input_add(sock, GDK_INPUT_WRITE,
					     network_conn_napigator_cb, NULL);
}

int network_download_start_timeout(gpointer data) {
   NAP_DOWNLOAD *download;
   
   if (!data) 
     return 0;
   
   download = (NAP_DOWNLOAD *)data;
   fprintf(stderr, "download_start_timeout ... status = %i\n", download->status);
   if (download->status < 2) 
     download_close(download, TRANSFER_ERROR);
   
   return 0;
}

void network_add_download(NAP_DOWNLOAD *download) {
   struct stat st;
   
   if (!download)
     return;
   
   download->upload = 0;
   gnapster_fill_dl_resume_path(download);
   if (download->resume_path) {
      if (stat(download->resume_path, &st) < 0) {
	 g_free(download->resume_path);
	 download->resume_path = NULL;
      }
   }
   
   gnapster_clist_append(download_tab.clist, download_tab.vbox,
			 download, download->trunc_file,
			 napster_conn_as_str(download->connection),
			 download->user, "Requesting...", NULL);
   
   /* 5 minutes... */
   download->start_timeout = gtk_timeout_add(60000 * 5, network_download_start_timeout,
					     download);
}

void network_start_download(NAP_DOWNLOAD *download) {
   if (!download)
     return;
   
   napster_send(conn_info.sock, NAPSTER_DOWNLOAD_START, NULL);
   
   download->calls = 5;
   download->status++;
   if (download->status != 2) {
      download_free(download);
      return;
   }
   
   if (!download->resume_path)
     append_resume_entry(download);
   
   gtk_timeout_remove(download->start_timeout);
   download->start_timeout = 0;
   download->update_tag = gtk_timeout_add(CLIST_UPDATE_SPEED, network_update_download,
					  download);
   download->sock_input = gdk_input_add(download->sock, GDK_INPUT_READ,
					network_handle_download, download);
}

void network_add_upload(NAP_DOWNLOAD *upload) {
   if (!upload)
     return;
   
   gnapster_text_insert(CONSOLE, SYSTEM, "Sending \2%s\1 to \2%s\1...\n",
			upload->trunc_file, upload->user);
   gnapster_clist_append(upload_tab.clist, upload_tab.vbox,
			 upload, upload->trunc_file,
			 napster_conn_as_str(upload->connection),
			 upload->user, "0.00M [0%] @ 0.0k/s", NULL);
   
   napster_send(conn_info.sock, NAPSTER_UPLOAD_START, NULL);
   
   upload->calls = 5;
   
   upload->update_tag = gtk_timeout_add(CLIST_UPDATE_SPEED, network_update_upload,
					upload);
   
   upload->sock_input = gdk_input_add(upload->sock, GDK_INPUT_WRITE,
				      network_handle_upload, upload);
}

void network_handle_submit(gpointer data, int source, GdkInputCondition condition) {
   GList *list, *link;
   MP3 *mp3;
   int x;

   if (!total_list) {
      autojoin_handle();
      total_list = mp3_list_free(total_list);
      gdk_input_remove(conn_info.upload_input);
      return;
   }
   
   list = g_list_nth(total_list, 0);
   if (!list || !conn_info.connected) {
      autojoin_handle();
      total_list = mp3_list_free(total_list);
      gdk_input_remove(conn_info.upload_input);
      return;
   }
   mp3 = (MP3 *)list->data;
   
   x = napster_send(conn_info.sock, NAPSTER_SUBMIT,
		    "\"%s\" %s %lu %li %lu %lu",
		    mp3->filename, mp3->checksum, mp3->filesize,
		    mp3->bitrate, mp3->freq, mp3->time);

   link = g_list_find(total_list, list->data);
   total_list = g_list_remove_link(total_list, link);
   g_free(mp3->filename);
   g_free(mp3->checksum);
   g_free(mp3);
   list->data = NULL;
      
   if (x < 0) {
      autojoin_handle();
      total_list = mp3_list_free(total_list);
      gdk_input_remove(conn_info.upload_input);
      return;
   }
}

void network_handle_sigpipe(int sig) {
   fprintf(stderr, "SIGPIPE received\n");
}

int network_fill_download(NAP_DOWNLOAD **download, char *user, char *file, int i, int upload) {
   char *tf;
   NAP_DOWNLOAD *dl;
   struct stat st;
   char *temp_str;
   
   dl = *download;
   if (!dl)
     return -1;
   
   dl->user = strdup(user);
   dl->file = strdup(file);
   if (upload)
     dl->size = i;
   else
     dl->total = i;
   
   tf = strrchr(file, '\\');
   if (!tf) {
      close(dl->sock);
      download_free(dl);
      return -1;
   }
   tf++;
   dl->trunc_file = strdup(tf);
   
   if (upload &&
       !upload_accepted(dl->user, dl->file, dl->trunc_file)) {
      close(dl->sock);
      download_free(dl);
      return -1;
   }

   if (upload) {
      /* Ok, this is a potential security flaw so lets check to make sure this
       * exists in ~/.gnapster/shared ... */
      if (!verify_file_shared(dl->file)) {
	 send(dl->sock, "FILE NOT SHARED", strlen("FILE NOT SHARED"), 0);
	 close(dl->sock);
	 download_free(dl);
	 return -1;
      }
      
      dl->fdesc = fopen(conv_unix(dl->file), "r");
   } else {
      if (!dl->resume_path)
	dl->fdesc = open_file(dl, dl->trunc_file);
      else {
	 dl->fdesc = fopen(dl->resume_path, "a");
	 dl->fpath = strdup(dl->resume_path);
      }
   }
   if (!dl->fdesc) {
      send(dl->sock, "FILE NOT FOUND", strlen("FILE NOT FOUND"), 0);
      close(dl->sock);
      download_free(dl);
      return -1;
   }
   
   if (upload) {
      if (dl->size)
	lseek(fileno(dl->fdesc), dl->size, SEEK_SET);
      fstat(fileno(dl->fdesc), &st);
//      dl->size = 0;
      dl->total = st.st_size;
      
      temp_str = g_strdup_printf("%lu", dl->total);
      send(dl->sock, temp_str, strlen(temp_str), 0);
      g_free(temp_str);
   } else {
      fstat(fileno(dl->fdesc), &st);
      dl->size = 0;
      if (st.st_size) {
	 dl->size = st.st_size;
	 gnapster_text_insert(CURR, SYSTEM, "Resume started at \2%s\1 for \2%s\1\n",
			      simplify_size(dl->size), dl->trunc_file);
//	 lseek(fileno(dl->fdesc), dl->size, SEEK_SET);
      }
      temp_str = g_strdup_printf("%lu", (unsigned long int)st.st_size);
      send(dl->sock, temp_str, strlen(temp_str), 0);
      g_free(temp_str);
      //dl->size = 0;
      dl->connection = 0;
   }
   
   return 1;
}

FILE *fc(FILE *f) {
   FILE *temp = NULL;
   
   memcpy(temp, f, sizeof(f));
   
   return temp;
}

void network_handle_header_complete(gpointer data, int source, GdkInputCondition condition) {
   char c;
   int n;
   NAP_DOWNLOAD *download;
   
   download = (NAP_DOWNLOAD *)data;
   
   if (download->flag == 3) {
      char last[16 + 1];
      char *ptr;
      char *user, *file;
      int iv;
      NAP_DOWNLOAD *real = NULL;
      
      gdk_input_remove(download->sock_input);
      gtk_timeout_remove(download->accept_timeout);

      download->header[download->header_val] = 0;
      n = recv(source, last, sizeof(last) - 1, 0);
      if (n)
	last[n] = 0;
      else {
	 close(download->sock);
	 download_free(download);
	 return;
      }
      ptr = download->header;
      user = next_arg(ptr, &ptr);
      file = next_arg(ptr, &ptr);
      if (!file) {
	 close(download->sock);
	 download_free(download);
	 return;
      }
      iv = my_atoi(last);
      
      if (!download->upload) {
	 real = gnapster_match_real_download(user, file, NULL);
	 if (!real) {
	    close(source);
	    download_free(download);
	    return;
	 }
	 if (real->resume_path)
	   download->resume_path = strdup(real->resume_path);
      }
      if (network_fill_download(&download, user, file, iv, download->upload) <= 0)
	return;
      if (download->upload)
	network_add_upload(download);
      else {
	 real->saddr = download->saddr;
	 real->sock = download->sock;
	 real->sock_input = download->sock_input;
	 real->fpath = strdup(download->fpath);
	 real->firewall = download->firewall;
	 real->upload = download->upload;
	 real->flag = download->flag;
	 real->user = strdup(download->user);
	 real->fdesc = download->fdesc;
	 real->size = download->size;
	 real->total = download->total;
	 real->ip = download->ip;
	 real->port = download->port;
	 real->file = strdup(download->file);
	 real->trunc_file = strdup(download->trunc_file);
	 download_free(download);
	 network_start_download(real);
      }
      return;
   }
   n = recv(source, &c, 1, 0);
   if (n > 0) {
      download->header[download->header_val++] = c;
      if (c == '"' || download->flag == 2)
	download->flag++;
   } else {
      close(download->sock);
      gdk_input_remove(download->sock_input);
      gtk_timeout_remove(download->accept_timeout);
      download_free(download);
   }
}

void network_handle_accept_header(gpointer data, int source, GdkInputCondition condition) {
   char c;
   int n;
   NAP_DOWNLOAD *download;
   
   download = (NAP_DOWNLOAD *)data;
   
   n = recv(source, &c, 1, 0);
   if (n <= 0 || download->header_val > 4) {
      fprintf(stderr, "download->header_val = %i\n",
	      download->header_val);
      close(download->sock);
      gdk_input_remove(download->sock_input);
      gtk_timeout_remove(download->accept_timeout);
      download_free(download);
      return;
   } else if (n > 0) {
      download->header[download->header_val++] = c;
      download->header[download->header_val] = '\0';
      if (!strcmp(download->header, "GET")) {
	 gdk_input_remove(download->sock_input);
	 bzero(download->header, sizeof(download->header));
	 download->header_val = 0;
	 download->flag = 0;
	 download->upload = 1;
	 download->sock_input = gdk_input_add(source, GDK_INPUT_READ,
					      network_handle_header_complete, 
					      download);
      } else if (!strcmp(download->header, "SEND")) {
	 gdk_input_remove(download->sock_input);
	 bzero(download->header, sizeof(download->header));
	 download->header_val = 0;
	 download->flag = 0;
	 download->upload = 0;
	 download->sock_input = gdk_input_add(source, GDK_INPUT_READ,
					      network_handle_header_complete, 
					      download);
      }
   }
}

int network_accept_timeout(gpointer data) {
   NAP_DOWNLOAD *download;
   
   if (!data) return 0;
   download = (NAP_DOWNLOAD *)data;
   
   fprintf(stderr, "Download timeout...\n");
   gdk_input_remove(download->sock_input);
   close(download->sock);
   download_free(download);
   
   return 0;
}

void network_handle_connection(gpointer data, int source, GdkInputCondition condition) {
   NAP_DOWNLOAD *download;
   struct sockaddr_in from;
   int len;
   
   download = g_malloc(sizeof(NAP_DOWNLOAD));
   memset(download, 0, sizeof(NAP_DOWNLOAD));
   download->type = N_DOWNLOAD;
   len = sizeof(from);
   download->sock = 
     accept(conn_info.bind_sock, (struct sockaddr *)&from, &len);
   if (download->sock < 0) {
      perror("accept()");
      download_free(download);
      return;
   }
   download->accept_timeout = 
     gtk_timeout_add(30000, network_accept_timeout, download);
   send(download->sock, "1", 1, 0);
   download->header_val = 0;
   
   download->sock_input = gdk_input_add(download->sock, GDK_INPUT_READ,
					network_handle_accept_header, download);
}

void download_close(NAP_DOWNLOAD *download, TRANSFER_INFO info) {
   char *temp_str = NULL;
   int incomplete = 0, row;

   if (!download)
     return;
   
   if (download->start_timeout)
     gtk_timeout_remove(download->start_timeout);
   if (download->sock_input)
     gdk_input_remove(download->sock_input);
   if (download->update_tag)
     gtk_timeout_remove(download->update_tag);
   if (download->timeout_tag)
     gtk_timeout_remove(download->timeout_tag);
   if (download->fdesc && download->status == 2)
     fclose(download->fdesc);
   
   if (info & TRANSFER_FILE_REMOVE && download->fpath && download->status == 2) {
      unlink(download->fpath);
      remove_resume_entry(download);
   }
   
   if (download->sock)
     close(download->sock);
   
   if (download->status == 2)
     napster_send(conn_info.sock, NAPSTER_DOWNLOAD_COMPLETE, NULL);

   if (download->status == 2) {
      if (info & TRANSFER_FINISHED) {
	 temp_str = g_strdup_printf("Download complete for \2%s\1\n",
				    download->trunc_file);
	 remove_resume_entry(download);
	 if (download->resume_path) {
	    char *path;
	    path = g_strdup_printf("%s/%s", user_info.download_dir,
				   download->trunc_file);
	    rename(download->fpath, path);
	 }
      } else if (info & TRANSFER_CLOSED) {
	 temp_str = g_strdup_printf("Download cancelled by \2%s\1 for \2%s\1 (\2%s\1 lost)...\n",
				    download->user, download->trunc_file, simplify_size(download->total - download->size));
	 incomplete = 1;
      } else if (info & TRANSFER_TIMEOUT) {
	 temp_str = g_strdup_printf("Download timed out for \2%s\1 from \2%s\1\n",
				    download->trunc_file, download->user);
	 incomplete = 1;
      } else if (info & TRANSFER_CANCEL) {
	 incomplete = 0;
      }
      if (info & TRANSFER_FILE_REMOVE)
	incomplete = 0;

      if (incomplete && download->fpath && !download->resume_path) {
	 char *path, *file;
	 char *new_path;

	 path = strdup(download->fpath);
	 file = strrchr(path, '/');
	 if (file) {
	    int x = -1;
	    struct stat st;

	    file[0] = '\0';
	    file++;
	    new_path = g_strdup_printf("%s/incomplete", user_info.download_dir);
	    if (stat(new_path, &st) < 0)
	      x = mkdir(new_path, 0755);
	    if (S_ISDIR(st.st_mode) || x >= 0) {
	       g_free(new_path);
	       /* Maybe we better use user_info.download_dir... */
	       new_path = g_strdup_printf("%s/incomplete/%s", 
					  user_info.download_dir, file);
	       rename(download->fpath, new_path);
	       /* Ok, now lets write it to the resume */
	       modify_resume_entry(new_path, download);
//	       append_resume_entry(new_path, download);
	    }
	    g_free(new_path);
	 }
	 g_free(path);
      }
   }
   
   if (temp_str) {
      gnapster_text_insert(CONSOLE, SYSTEM, temp_str);
      g_free(temp_str);
   }

   row = gtk_clist_find_row_from_data(GTK_CLIST(download_tab.clist), download);
   if (row >= 0)
     gtk_clist_remove(GTK_CLIST(download_tab.clist), row);
   
   queue_check();
}

void network_handle_download(gpointer data, int source, GdkInputCondition condition) {
   char buf[2048];
   int n;
   NAP_DOWNLOAD *download;
   
   download = (NAP_DOWNLOAD *)data;
   
   bzero(buf, sizeof(buf));
   n = recv(source, buf, 2048, 0);
   if (n <= 0) {
      /* Could someone explain to me why this occurs? */
      if (download->size == download->total || 
	  (download->total - download->size) <= 2048)
	download_close(download, TRANSFER_FINISHED);
      else
	download_close(download, TRANSFER_CLOSED);

      return;
   } else if (n > 0) {
      if (fwrite(buf, n, sizeof(char), download->fdesc) <= 0) {
	 gnapster_text_insert(CONSOLE, SYSTEM, "Error writing to disk!\n");
	 download_close(download, TRANSFER_CANCEL);
	 return;
      }
      download->size += n;
      download->temp_size += n;
   }
}

int network_update_download(gpointer data) {
   char *temp_str;
   NAP_DOWNLOAD *download;
   
   if (!data) 
     return 0;   
   download = (NAP_DOWNLOAD *)data;
   
   download->calls++;
   if (download->calls >= 3) { /* x seconds */
      if (download->temp_size == 0)
	download->timeout_calls++;
      else
	download->timeout_calls = 0;

      download->kps = 
	(((float)download->temp_size / (float)1024)) / (float)(CLIST_UPDATE_SPEED / 1000);
      download->temp_size = 0;
      download->calls = 0;
      
      if (download->timeout_calls == ((5 * 60) / 3)) {
	 download_close(download, TRANSFER_TIMEOUT);
	 return 0;
      }
   }
   
   //seconds = (download->size / ((download->kbps * 1000) / 8));
   temp_str = g_strdup_printf("%.02fM [%i%%] @ %.01fk/s",
			      ((float)download->size / 1024) / 1024,
			      (int)((float)download->size / (float)download->total * 100),
			      (float)download->kps);

   gtk_clist_set_text(GTK_CLIST(download_tab.clist),
		      gtk_clist_find_row_from_data(GTK_CLIST(download_tab.clist),
						   download),
		      3, temp_str);
   g_free(temp_str);
   download->temp_size = 0;

   return 1;
}

int network_update_upload(gpointer data) {
   char *temp_str;
   NAP_DOWNLOAD *upload;
   
   if (!data) return 0;   
   upload = (NAP_DOWNLOAD *)data;
   
   upload->calls++;
   if (upload->calls >= 3) { /* x seconds */
      if (upload->temp_size == 0)
	upload->timeout_calls++;
      else
	upload->timeout_calls = 0;

      upload->kps =
	(((float)upload->temp_size / (float)1024)) / 
	((float)(CLIST_UPDATE_SPEED / 1000) * 3);
      upload->temp_size = 0;
      upload->calls = 0;
     
      if (upload->timeout_calls >= ((5 * 60) / 3)) {
	 upload_close(upload, TRANSFER_TIMEOUT);
	 return 0;
      }
   }
   
   temp_str = g_strdup_printf("%.02fM [%i%%] @ %.01fk/s",
			      ((float)upload->size / 1024) / 1024,
			      (int)((float)upload->size / (float)upload->total * 100),
			      (float)upload->kps);
   
   gtk_clist_set_text(GTK_CLIST(upload_tab.clist),
		      gtk_clist_find_row_from_data(GTK_CLIST(upload_tab.clist),
						   upload),
		      3, temp_str);
   g_free(temp_str);
//   upload->temp_size = 0;
   
   return 1;
}

void upload_close(NAP_DOWNLOAD *upload, TRANSFER_INFO info) {
   char *temp_str = NULL;
	 
   gdk_input_remove(upload->sock_input);
   gtk_timeout_remove(upload->update_tag);

   napster_send(conn_info.sock, NAPSTER_UPLOAD_COMPLETE, NULL);
   
   if (info & TRANSFER_CLOSED)
     temp_str = g_strdup_printf("File upload cancelled for \2%s\1 by \2%s\1 (\2%s\1 lost)\n",
				upload->trunc_file, upload->user, simplify_size(upload->total - upload->size));
   else if (info & TRANSFER_FINISHED)
     temp_str = g_strdup_printf("File upload complete for \2%s\1 to \2%s\1\n",
				upload->trunc_file, upload->user);
   else if (info & TRANSFER_TIMEOUT)
     temp_str = g_strdup_printf("File upload timed out for \2%s\1 to \2%s\1\n",
				upload->trunc_file, upload->user);
   
   if (temp_str) {
      gnapster_text_insert(CONSOLE, SYSTEM, temp_str);
      g_free(temp_str);
   }
   
   close(upload->sock);
   fclose(upload->fdesc);
   gtk_clist_remove(GTK_CLIST(upload_tab.clist),
		    gtk_clist_find_row_from_data(GTK_CLIST(upload_tab.clist), upload));
}

void network_handle_upload(gpointer data, int source, GdkInputCondition condition) {
   char buf[2048 + 1];
   int n, sent;
   NAP_DOWNLOAD *upload;
   
   upload = (NAP_DOWNLOAD *)data;
   if (!upload) return;
   
   if (upload->size < upload->total) {
      lseek(fileno(upload->fdesc), upload->size, SEEK_SET);
      n = read(fileno(upload->fdesc), buf, 1024);
      if (n <= 0) {
	 upload_close(upload, TRANSFER_CLOSED);
	 return;
      }
      sent = send(upload->sock, buf, n, 0);
   
      if (sent <= 0) {
	 upload_close(upload, TRANSFER_CLOSED);
      } else {
	 upload->size += sent;
	 upload->temp_size += sent;
      }
   } else
     upload_close(upload, TRANSFER_FINISHED);
}

void network_conn_user_header_cb(gpointer data, int source, GdkInputCondition condition) {
   NAP_DOWNLOAD *download;
   char c;
   int n;
   
   download = (NAP_DOWNLOAD *)data;
   if (!download) return;

   n = recv(source, &c, 1, 0);
   if (download->firewall) {
      /* Ok lets transfer this code over to the true agent now */
      gdk_input_remove(download->sock_input);
      if (c != '1')
	return;
      n = recv(source, &c, 1, 0);
      if (n <= 0 || !isdigit((unsigned char)c))
	return;

      network_add_upload(download);
      return;
   }
   
   if (n <= 0) {
      download->header[download->header_val] = '\0';
      gnapster_text_insert(CONSOLE, SYSTEM, "Connection closed by remote user \2%s\1\n",
			   download->user);
      download_close(download, TRANSFER_ERROR);
      return;
   }

   if (c == '\n') return;
   
   if (!isdigit((unsigned char)c)) {
      char *temp_str;
      
      download->header[download->header_val] = '\0';
      gdk_input_remove(download->sock_input);
      
      if (!strcmp(download->header, "1") || download->header_val <= 1) {
	 gnapster_text_insert(CONSOLE, SYSTEM, "Could not access file from \2%s\1 (try again?)\n",
			      download->user);
	 download_close(download, TRANSFER_ERROR);
	 return;	 
      }
      
      temp_str = download->header;
      temp_str++;
      
      download->total = my_atoi(temp_str);
      
      if (!download->resume_path)
	download->fdesc = open_file(download, download->trunc_file);
/*      else
	download->fdesc = fopen(download->resume_path, "a");*/
      if (!download->fdesc) {
	 gnome_error_dialog(_("Uh oh.  The path you've chosen does not exist or you do not have write permissions\nPlease open up napster's preferences and pick a directory you know you have access to"));
	 download_close(download, TRANSFER_ERROR);
/*	 close(source);
	 download_free(download);*/
	 return;
      }
      
//      fstat(fileno(download->fdesc), &st);
      if (!download->size)
	download->size = 1;
      else
	download->size++;
/*      if (st.st_size) {
	 download->size = st.st_size + 1;
	 lseek(fileno(download->fdesc), download->size - 1, SEEK_SET);
      }*/
      if (fwrite(&c, 1, sizeof(char), download->fdesc) <= 0) {
	 gnapster_text_insert(CONSOLE, SYSTEM, "Error writing to disk!\n");
	 download_close(download, TRANSFER_CANCEL);
	 return;
      }
      network_start_download(download);
   } else {
      download->header[download->header_val++] = c;
   }
}

void network_conn_user_cb(gpointer data, int source, GdkInputCondition condition) {
   NAP_DOWNLOAD *download;
   int err, len;
   
   download = (NAP_DOWNLOAD *)data;

   gdk_input_remove(download->sock_input);

   len = sizeof(err);
   getsockopt(source, SOL_SOCKET, SO_ERROR, &err, &len);
   if (err) {
      gnapster_text_insert(CONSOLE, SYSTEM, "Error connecting to user...\n");
      download_close(download, TRANSFER_ERROR);
      return;
   }

   fcntl(source, F_SETFL, 0);   
   
   if (!download->firewall) {
      if (download->resume_path) {
	 struct stat st;
	 download->fdesc = fopen(download->resume_path, "a");
	 download->fpath = strdup(download->resume_path);
	 fstat(fileno(download->fdesc), &st);
	 if (st.st_size) {
	    download->size = st.st_size;
//	    lseek(fileno(download->fdesc), download->size, SEEK_SET);
	 }
	 gnapster_text_insert(CURR, SYSTEM, "Resume started at \2%s\1 for \2%s\1\n", 
			      simplify_size(download->size), 
			      download->trunc_file);
      } else
	download->size = 0;
      napster_send_get(source, user_info.user, download->file,
		       download->size);
   } else {
      struct stat st;

      if (!upload_accepted(download->user, download->file, download->trunc_file)) {
	 close(download->sock);
	 download_free(download);
	 return;
      }

      if (!verify_file_shared(download->file)) {
	 send(download->sock, "FILE NOT SHARED", strlen("FILE NOT SHARED"), 0);
	 close(download->sock);
	 download_free(download);
	 return;
      }
      download->fdesc = fopen(conv_unix(download->file), "r");
      if (!download->fdesc) {
	 char *temp_str;
	 
	 temp_str = g_strdup_printf("FILE NOT FOUND");
	 send(download->sock, temp_str, strlen(temp_str), 0);
	 g_free(temp_str);
	 close(download->sock);
	 download_free(download);
	 return;
      }

      fstat(fileno(download->fdesc), &st);
      download->size = 0;
      download->total = st.st_size;

      conv_win(download->file);
      napster_send_send(source, user_info.user, download->file,
			download->total);
   }

   download->sock_input = gdk_input_add(source, GDK_INPUT_READ,
					network_conn_user_header_cb, download);
}

void network_read_best_host_cb(gpointer data, int source, GdkInputCondition condition) {
   char buf[1024];
   char *ptr, *ip;
   int port, n;
   
   n = recv(source, buf, sizeof(buf) - 1, 0);
   if (n <= 0) {
      perror("recv");
      disconnect_cb();
      return;
   }
   buf[n] = '\0';
   ptr = buf;
   
   ip = next_arg_full(ptr, &ptr, ':');
   port = my_atoi(next_arg_full(ptr, &ptr, ':'));
   
   g_free(conn_info.ip);
   conn_info.ip = strdup(ip);
   conn_info.port = port;
   
   gdk_input_remove(conn_info.sock_input);
   close(source);

   if (!strcmp(conn_info.ip, "127.0.0.1")) {
      gnapster_update_status(_("Napster query server reported all servers were busy"), 0);
      conn_info.connected = 0;
      conn_info.connecting = 0;
      gtk_widget_set_sensitive(search_tab.media, 0);
      return;
   }
   
   gnapster_update_status(_("Best host found, connecting..."), -1);
   napster_connect(conn_info.ip, conn_info.port);   
}

void network_conn_cb(gpointer data, int source, GdkInputCondition condition) {
   int err, len;
 
   gdk_input_remove(conn_info.sock_input);

   len = sizeof(err);
   getsockopt(source, SOL_SOCKET, SO_ERROR, &err, &len);
   if (err) {
      disconnect_cb();
/*      close(source);
      gnapster_update_status(_("Unknown fatal error connecting to server (try again?)"), 0);
      conn_info.connected = 0;
      conn_info.connecting = 0;
      gtk_widget_set_sensitive(search_tab.media, 0);*/
      return;
   }

   fcntl(source, F_SETFL, 0);
//   gdk_input_remove(conn_info.sock_input);
   conn_info.sock_input = gdk_input_add(source, GDK_INPUT_READ,
					network_read_best_host_cb, NULL);
}

void network_read_cb(gpointer data, int source, GdkInputCondition condition) {
   int n;
   unsigned char buf[355];
   unsigned char buf_holder[355];
   int old_n, i;
   PACKET *p;

   if (read_count == 0) {
      total_packet_len = 4;
      packet_len = 0;
      read_count++;
   }

   if (read_count == 1) {
      n = 0;
      bzero(buf, sizeof(buf));
      while(n < 4) {
	 bzero(buf_holder, sizeof(buf_holder));
	 old_n = n;
	 n += recv(source, buf_holder, (4 - n), 0);
	 if (n <= old_n) {
	    n = 0;
	    perror("recv");
	    disconnect_cb();
	    return;
	 }
	 for(i=0; i<(n-old_n); i++)
	   buf[old_n + i] = buf_holder[i];
      }
      
      p = g_malloc(sizeof(PACKET));
      memcpy(p, buf, 4);
      p->len = BSWAP16(p->len);
      p->cmd = BSWAP16(p->cmd);
      DPRINTF(("l=%i,c=%i\n", p->len, p->cmd));

      bzero(packet_data, sizeof(packet_data));
      
      packet_len = 0;
      total_packet_len = (int)p->len;
      packet_cmd = (int)p->cmd;
      g_free(p);
      
      read_count++;
      if (packet_len == total_packet_len) read_count++;
   } else if (read_count == 2) {
      int lenbuf;
      bzero(buf_holder, sizeof(buf_holder));
      lenbuf = recv(source, buf_holder, 
		    ((total_packet_len - packet_len) > sizeof(buf_holder)) ? 
		    sizeof(buf_holder) : (total_packet_len - packet_len),
		    0);
      if (lenbuf < 0) {
	 perror("recv");
	 disconnect_cb();
	 return;
      }
      packet_len += lenbuf;
      /* If this fails we'll pretty much just ignore the rest of the data... */
      if (packet_len < 4096)
	strcat(packet_data, buf_holder);
      
      if (packet_len >= total_packet_len) read_count++;
   }
   
   if (read_count == 3) {
      DPRINTF(("d=%s\n", packet_data));

      if (!conn_info.connected) {
	 conn_info.connecting = 0;
	 conn_info.connected = 1;
      }

      napster_handle_data(packet_cmd, packet_data);
      
      read_count = 0;
   }
}

void network_conn_real_cb(gpointer data, int source, GdkInputCondition condition) {
   char temp_str[80];
   int err, len;
   
   gdk_input_remove(conn_info.sock_input);

   len = sizeof(err);
   getsockopt(source, SOL_SOCKET, SO_ERROR, &err, &len);
   if (err) {
      gnapster_text_insert(CONSOLE, SYSTEM, "Failure connecting to napster server.\n");
      disconnect_cb();
/*      gnapster_update_status(_("Failed to connect"), 0);
      conn_info.connecting = conn_info.connected = 0;
      gtk_widget_set_sensitive(search_tab.media, 0);*/
      return;
   }

   fcntl(source, F_SETFL, 0);

   if (user_info.new_user)
     napster_send(source, NAPSTER_NEW_USER, user_info.user);
   napster_send(source, NAPSTER_LOGIN, "%s %s %i \"%s\" %i",
		user_info.user, user_info.pass, conn_info.nap_port, NAP_VERSION,
		user_info.connection);
   
   g_snprintf(temp_str, sizeof(temp_str), _("Connected (%s:%i)...awaiting login reply..."),
	      conn_info.ip, conn_info.port);
   gnapster_update_status(temp_str, -1);
   conn_info.sock_input = gdk_input_add(source, GDK_INPUT_READ,
					network_read_cb, NULL);
}
