#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>

#include <gnome.h>

#include "napster.h"
#include "gnapster.h"
#include "callbacks.h"
#include "ui.h"
#include "upload.h"
#include "network.h"
#include "commands.h"
#include "download.h"
#include "queue.h"
#include "resume.h"

#include "config.h"

extern RESUME_DLG resume_dlg;
extern CONN_INFO conn_info;
extern SEARCH_TAB search_tab;
extern BROWSE_DLG browse_dlg;

void modify_resume_entry(char *newpath, NAP_DOWNLOAD *download) {
   FILE *f;
   char *hld, *file, *size, *md5, *dptr;
   GList *ptr, *felem;
   
   f = open_resume("r");
   if (!f)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   f = open_resume("w");
   if (!f)
     return;
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      file = next_arg_first(dptr, &dptr, &hld);
      size = next_arg(dptr, &dptr);
      md5 = next_arg(dptr, &dptr);
      if (md5) {
	 if (!strcmp(md5, download->id))
	   fprintf(f, "\"%s\" %i %s\n", newpath, download->size, md5);
	 else
	   fprintf(f, "\"%s\" %s %s\n", file, size, md5);
      }
      
      g_free(hld);
   }
   
   g_list_free(felem);
   
   fclose(f);
}

void remove_resume_entry(NAP_DOWNLOAD *download) {
   FILE *f;
   char *hld, *file, *size, *md5, *dptr;
   GList *ptr, *felem;
   
   f = open_resume("r");
   if (!f)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   f = open_resume("w");
   if (!f)
     return;
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      file = next_arg_first(dptr, &dptr, &hld);
      size = next_arg(dptr, &dptr);
      md5 = next_arg(dptr, &dptr);
      
      if (md5) {
	 if (strcmp(md5, download->id))
	   fprintf(f, "\"%s\" %s %s", file, size, md5);
      }
      
      g_free(hld);
   }
   
   g_list_free(felem);
   
   fclose(f);
}

void append_resume_entry(NAP_DOWNLOAD *download) {
   FILE *f;
   char *hld, *file, *size, *md5, *dptr;
   GList *ptr, *felem;
   
   f = open_resume("r");
   if (!f)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      file = next_arg_first(dptr, &dptr, &hld);
      size = next_arg(dptr, &dptr);
      md5 = next_arg(dptr, &dptr);
      
      if (md5) {
	 if (!strcmp(md5, download->id))
	   break;
      }
      
      g_free(hld);
   }
   
   g_list_free(felem);
   
   f = open_resume("a");
   if (!f)
     return;
   
   if (!strchr(download->id, '-'))
     fprintf(f, "\"%s\" %i %s-%li\n", download->fpath, download->size, 
	     download->id, download->total);
   else
     fprintf(f, "\"%s\" %i %s\n", download->fpath, download->size,
	     download->id);
   
   fclose(f);
}

void resume_dlg_clist_cb() {
}

void resume_dlg_cb(GtkWidget *w, gpointer data) {
   char *s;
   
   if (!data)
     return;
   
   s = (char *)data;
   
   if (!strcmp(s, "resume")) {
      /* Resume searches are actually broken server side and on the client-side
       * of the official client...so kind of useless */
      NAP_RESUME *temp;
      
      temp = gnapster_get_selection_data(resume_dlg.clist, NULL);
      if (!temp)
	return;
      
      napster_send(conn_info.sock, NAPSTER_RESUME_SEARCH, "%s-%i %i",
		   temp->checksum, temp->remotesize, temp->remotesize);
      
      conn_info.searching = 1;
      
      gtk_clist_clear(GTK_CLIST(search_tab.clist));
      gtk_clist_freeze(GTK_CLIST(search_tab.clist));
      gtk_entry_set_editable(GTK_ENTRY(search_tab.gtk_entry), 0);
   } else if (!strcmp(s, "remove")) {
      NAP_RESUME *temp;
      NAP_DOWNLOAD *download;
      
      temp = gnapster_get_selection_data(resume_dlg.clist, NULL);
      if (!temp)
	return;
      
      download = g_malloc(sizeof(NAP_DOWNLOAD));
      memset(download, 0, sizeof(NAP_DOWNLOAD));
      
      download->type = N_DOWNLOAD;
      download->id = g_strdup_printf("%s-%i", temp->checksum,
				     temp->remotesize);
      
      remove_resume_entry(download);
      
      g_free(download);
      
      populate_resume_dlg();
      
      return;
   }
   
   gnome_dialog_close(GNOME_DIALOG(resume_dlg.window));
   resume_dlg.window = NULL;
}

void populate_resume_dlg() {
   FILE *f;
   char *hld, *file, *local, *remote, *tf, *dptr;
   GList *ptr, *felem;
   NAP_RESUME *resume;
   
   f = open_resume("r");
   if (!f)
     return;
   
   gtk_clist_freeze(GTK_CLIST(resume_dlg.clist));
   gtk_clist_clear(GTK_CLIST(resume_dlg.clist));
   
   felem = read_file(f);
   
   fclose(f);
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      file = next_arg_first(dptr, &dptr, &hld);
      local = next_arg(dptr, &dptr);
      remote = next_arg(dptr, &dptr);
      if (!remote)
	break;
      
      resume = g_malloc(sizeof(NAP_RESUME));
      memset(resume, 0, sizeof(NAP_RESUME));
      
      resume->type = N_RESUME;
      resume->file = strdup(file);
      resume->localsize = my_atoi(local);
      resume->checksum = g_strdup(next_arg_full(remote, &remote, '-'));
      resume->remotesize = my_atoi(remote);
      
      tf = strrchr(resume->file, '/');
      if (tf)
	tf++;
      
      gnapster_clist_append(resume_dlg.clist, NULL, resume, tf, 
			    simplify_size(resume->remotesize), NULL);
      
      g_free(hld);
   }
   
   g_list_free(felem);

   gtk_clist_thaw(GTK_CLIST(resume_dlg.clist));
}

void resume_download_cb() {
   if (resume_dlg.window)
     return;
   
   create_resume_dlg();
   populate_resume_dlg();
}

FILE *open_resume(char *flags) {
   FILE *f;
   char *conf;
   
   conf = g_strdup_printf("%s/.gnapster/resume", getenv("HOME"));
   
   f = fopen(conf, flags);
   
   if (!f)
     perror("fopen");
   
   g_free(conf);
   
   return f;
}
