#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>

#include <gnome.h>

#include "napster.h"
#include "gnapster.h"
#include "callbacks.h"
#include "ui.h"
#include "upload.h"
#include "network.h"
#include "commands.h"
#include "download.h"
#include "queue.h"

#include "servers.h"

#include "config.h"

extern BROWSE_DLG browse_dlg;
extern ADD_DLG add_dlg;
extern EDIT_DLG edit_dlg;
extern CONN_INFO conn_info;
extern REFRESH_DLG refresh_dlg;

void browse_opennap_cb() {
   if (browse_dlg.window && !GTK_WIDGET_VISIBLE(browse_dlg.window)) return;
   create_browse_dlg();
}

void browse_dlg_connect_cb(GtkWidget *w, void *data) {
   GtkCList *clist;
   NAP_SERVER *nsrv;

   clist = GTK_CLIST(browse_dlg.clist);

   if (w) {
      int *v = (int *)data;
      if (v)
	nsrv = gtk_clist_get_row_data(clist, *v);
      else
	nsrv = gnapster_get_selection_data(GTK_WIDGET(clist), NULL);
      if (!nsrv) return;

      if (conn_info.connected || conn_info.connecting) {
	 already_connected(nsrv);
	 return;
      }
   } else
     nsrv = (NAP_SERVER *)data;

   if (!nsrv)
     return;

   g_free(conn_info.ip);
   conn_info.ip = strdup(nsrv->ip);
   conn_info.port = nsrv->port;
   conn_info.connecting = 1;
   conn_info.connected = 0;

   gnapster_update_status(_("Connecting to server..."), 1);
   gnome_config_set_string("/gnapster/Options/server", conn_info.ip);
   gnome_config_set_int("/gnapster/Options/server_port", conn_info.port);
   gnome_config_sync();
   napster_connect(conn_info.ip, conn_info.port);

   browse_dlg_cancel_cb();
}

void browse_dlg_cancel_cb() {
   gtk_clist_clear(GTK_CLIST(browse_dlg.clist));
   gnome_dialog_close(GNOME_DIALOG(browse_dlg.window));
   browse_dlg.window = NULL;
}

void browse_dlg_add() {
   create_add_dlg();
   gnome_dialog_run(GNOME_DIALOG(add_dlg.window));
}

void browse_dlg_edit() {
   NAP_SERVER *nsrv;
   int row = -1;
   char *desc;

   nsrv = gnapster_get_selection_data(browse_dlg.clist, &row);
   if (!nsrv || row < 0) return;

   gtk_clist_get_text(GTK_CLIST(browse_dlg.clist), row, 0, &desc);
   if (!desc) return;
   if (!strlen(desc)) return;

   create_edit_dlg(nsrv, desc);
   gnome_dialog_run(GNOME_DIALOG(edit_dlg.window));
}

void browse_dlg_delete(GtkWidget *w, int *data) {
   FILE *f;
   char *ipport, *dptr, cmp[1024];
   GList *ptr, *felem;
   NAP_SERVER *nsrv;
   
   f = open_servers("r");
   if (!f)
     return;
   
   if (data)
     nsrv = gtk_clist_get_row_data(GTK_CLIST(browse_dlg.clist), *data);
   else
     nsrv = gnapster_get_selection_data(browse_dlg.clist, NULL);
   
   if (!nsrv)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   f = open_servers("w");
   
   g_snprintf(cmp, sizeof(cmp), "%s:%i",
	      nsrv->ip, nsrv->port);
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      ipport = next_arg(dptr, &dptr);
      if (!ipport) 
	break;
      
      if (strcmp(ipport, cmp))
	fprintf(f, "%s %s\n", ipport, next_arg(dptr, &dptr));
      
      g_free(ipport);
   }
   
   fclose(f);

   g_list_free(felem);
   
   if (!data)
     populate_browse_dlg_clist();
}

void populate_browse_dlg_clist() {
   FILE *f;
   char *ip, *port, *title, *dptr;
   GList *ptr, *felem;
   NAP_SERVER *nsrv;
   
   f = open_servers("r");
   if (!f)
     return;
   
   gtk_clist_freeze(GTK_CLIST(browse_dlg.clist));
   gtk_clist_clear(GTK_CLIST(browse_dlg.clist));
   
   felem = read_file(f);
   
   fclose(f);
   
   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      ip = next_arg_full(dptr, &dptr, ':');
      port = next_arg(dptr, &dptr);
      title = last_arg(dptr, &dptr);
      if (!title)
	continue;
      
      nsrv = g_malloc(sizeof(NAP_SERVER));
      memset(nsrv, 0, sizeof(NAP_SERVER));
      
      nsrv->type = N_SERVER;
      nsrv->ip = strdup(ip);
      nsrv->port = my_atoi(port);
      nsrv->desc = strdup(title);
      
      gnapster_clist_append(browse_dlg.clist, NULL, nsrv, nsrv->desc,
			    nsrv->ip, NULL);
      
      g_free(ip);
   }
   
   g_list_free(felem);
   
   gtk_clist_thaw(GTK_CLIST(browse_dlg.clist));
}

void browse_dlg_button_cb(GtkWidget *widget, GdkEventButton *event, gpointer data) {
   int row = -1, col;

   gtk_clist_get_selection_info(GTK_CLIST(widget), event->x, event->y,
				&row, &col);

   if (event->type == GDK_2BUTTON_PRESS && event->button == 1) {
      browse_dlg_connect_cb(widget, &row);
   }
}

void browse_dlg_refresh(GtkWidget *w, char *data) {
   network_connect_napigator();
   create_refresh_dlg();
   gnome_dialog_run(GNOME_DIALOG(refresh_dlg.window));
}

void refresh_dlg_cb(GtkWidget *w, void *data) {
   if (conn_info.napigator_input)
     gdk_input_remove(conn_info.napigator_input);
   if (conn_info.napigator_sock)
     close(conn_info.napigator_sock);
   
   gnome_dialog_close(GNOME_DIALOG(refresh_dlg.window));   
}

void add_dlg_cb(GtkWidget *w, char *data) {
   if (!data)
     return;
   
   if (!strcmp(data, "ok")) {
      char *server, *port, *title;
      
      server = gtk_entry_get_text(GTK_ENTRY(add_dlg.server));
      port = gtk_entry_get_text(GTK_ENTRY(add_dlg.port));
      title = gtk_entry_get_text(GTK_ENTRY(add_dlg.desc));
      if (!strlen(server) || !strlen(port) || !strlen(title))
	return;
      
      append_server_entry(server, port, title);
      
      populate_browse_dlg_clist();
   }
   
   gnome_dialog_close(GNOME_DIALOG(add_dlg.window));
}

void edit_dlg_cb(GtkWidget *w, NAP_SERVER *data) {
   if (data) {
      char *server, *port, *title;
      
      server = gtk_entry_get_text(GTK_ENTRY(edit_dlg.server));
      port = gtk_entry_get_text(GTK_ENTRY(edit_dlg.port));
      title = gtk_entry_get_text(GTK_ENTRY(edit_dlg.desc));
      if (!strlen(server) || !strlen(port) || !strlen(title))
	return;
      
      modify_server_entry(data->ip, data->port, server, port, title);
      
      populate_browse_dlg_clist();
   }
   
   gnome_dialog_close(GNOME_DIALOG(edit_dlg.window));
}

FILE *open_servers(char *flags) {
   FILE *f;
   char *conf;

   conf = g_strdup_printf("%s/.gnapster/servers", getenv("HOME"));
   
   f = fopen(conf, flags);

   if (!f)
     perror("fopen");

   g_free(conf);
   
   return f;
}

void modify_server_entry(char *oip, unsigned short int oport, char *ip, char *port, char *title) {
   FILE *f;
   char *lip, *lport, *ltitle, *dptr;
   GList *ptr, *felem;
   
   f = open_servers("r");
   if (!f)
     return;
   
   felem = read_file(f);
   
   fclose(f);
   
   f = open_servers("w");

   for(ptr=felem; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      
      lip = next_arg_full(dptr, &dptr, ':');
      lport = next_arg(dptr, &dptr);
      ltitle = next_arg(dptr, &dptr);
      if (!ltitle) {
	 if (lip) g_free(lip);
	 else g_free(dptr);
	 
	 continue;
      }
      
      if (!strcmp(oip, lip) && oport == my_atoi(lport))
	fprintf(f, "%s:%s %s\n", ip, port, title);
      else
	fprintf(f, "%s:%s %s\n", lip, lport, ltitle);
      
      g_free(lip);
   }
   
   g_list_free(felem);
   
   fclose(f);
}

void append_server_entry(char *ip, char *port, char *server) {
   FILE *f;
   
   f = open_servers("a");
   if (!f)
     return;

   fprintf(f,"%s:%s %s\n",
	   ip, port, server);
   
   fclose(f);
}

int check_server_match(char *ip, char *port) {
   FILE *f;
   
   char buf[1024], *ptr;
   char *mip, *mport;
   
   f = open_servers("r");
   if (!f)
     return 0;
   
   while(!feof(f)) {
      fgets(buf, sizeof(buf) - 1, f);
      if (feof(f))
	break;
      
      ptr = buf;
      
      mip = next_arg_full(ptr, &ptr, ':');
      mport = next_arg(ptr, &ptr);
      
      if ((ip && port) && (!strcmp(ip, mip) && !strcmp(port, mport)))
        return 1;
   }
   
   fclose(f);
   
   return 0;
}

